# util.py - kantag miscellaneous shared functionality.
# Copyright (C) 2018 David Gasaway
# https://bitbucket.org/dgasaway/kantag/

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
import collections
import argparse
import re


class ToggleAction(argparse.Action):
    """
    An argparse Action that will convert a y/n arg value into True/False.
    """
    def __call__(self, parser, ns, values, option_string):
        setattr(ns, self.dest, True if values[0] == 'y' else False)

""" Basic metadata pair. """
TagValue = collections.namedtuple('TagValue', 'tag, value')
""" Artist name and performance role. """
NameRole = collections.namedtuple('NameRole', 'name, role')
""" Work title and list of part titles. """
WorkParts = collections.namedtuple('WorkParts', 'work, parts')
""" Constituent parts of an old-style MusicBrainz release title. """
AlbumTitle = collections.namedtuple('AlbumTitle', 'title, discnum, subtitle')

# --------------------------------------------------------------------------------------------------
def parse_artist_role(artist):
    """
    Parse a musicbrainz-style performer name/role string into parts.  The result is a NamedRole
    named tuple.
    """
    match = re.match(r'^(?P<name>[^(]*) \((?P<role>.*)\)$', artist)
    if match:
        return NameRole(match.group('name'), match.group('role'))
    else:
        return NameRole(artist, None)

# --------------------------------------------------------------------------------------------------
def remove_artist_roles(artists):
    """
    Return a list of artists with the musicbrainz-style artist roles removed.  The argument should
    be a list of strings.
    """
    return [parse_artist_role(artist).name for artist in artists]

# --------------------------------------------------------------------------------------------------
def parse_track_title(full_title):
    """
    Parse a musicbrainz-style track title into work and parts.  The result is a WorkParts named
    tuple.
    """
    work = None
    match = re.match(r'^(?P<work>[^:]+): (?P<parts>.+)$', full_title)
    if match:
        work = match.group('work')
        s = match.group('parts')
    else:
        s = full_title

    parts = None
    if s.find(' / ') > -1:
        parts = s.split(' / ')
    elif work is not None:
        parts = [s]

    return WorkParts(work, parts)

# --------------------------------------------------------------------------------------------------
def parse_album_title(full_title):
    """
    Parse a musicbrainz-style release title into a tuple containing the release title, the disc
    number, and the disc subtitle.  The result is an AlbumTitle named tuple.
    """
    match = re.match(r'(?P<title>.+) \(disc (?P<disc>\d+)(: (?P<subtitle>.*))?\)', full_title)
    if match:
        return AlbumTitle(match.group('title'), match.group('disc'), match.group('subtitle'))
    else:
        return AlbumTitle(full_title, None, None)

# --------------------------------------------------------------------------------------------------
def expand_ranges(range_str):
    """
    Expand a range string into a list of individual values, e.g., '01-03,05' -> ['01', '02, '03',
    '05'].  Each number in the string should have the same number of digits.
    """
    if not re.match(r'^(\d+(-\d+)?)(,(\d+(-\d+)?))*$', range_str):
        raise exceptions.TagFileFormatError('Malformed track number range string: ' + range_str)

    result = []
    for item in range_str.split(','):
        match = re.match(r'(?P<start>\d+)-(?P<end>\d+)', item)
        if match:
            # Store the length of the start value so we can pad all values to
            # this length.
            item_len = len(match.group('start'))

            # Loop through the range, and add add each item to the list.
            start = int(match.group('start'))
            end = int(match.group('end'))
            for i in range(start, end+1):
                result.append(str(i).zfill(item_len))
        else:
            result.append(item)

    return result

# --------------------------------------------------------------------------------------------------
def condense_ranges(nums):
    """
    Given a list of numeric strings, sort and condense consecutive ranges into a range string,
    e.g., ['01', '02', '03', '05'] -> '01-03,05'.
    """
    # 'base' is the index storing the low value of a range.
    base = 0
    ranges = []
    nums = sorted([num for num in nums if not num is None])
    for idx in range(len(nums)):
        # If last index, or there is a gap before the next value, then output a range.
        if (idx == len(nums) - 1) or (int(nums[idx + 1]) - int(nums[idx]) > 1):
            ranges.append(('' if idx == base else str(nums[base]) + '-') + str(nums[idx]))
            # Next value will start the next range.
            base = idx + 1

    return ','.join(ranges)




# exceptions.py - kantag exceptions.
# Copyright (C) 2018 David Gasaway
# https://bitbucket.org/dgasaway/kantag/

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
class TaggingError(Exception): pass
class ProgramCallError(TaggingError): pass
class FileTypeError(TaggingError): pass
class FilenameError(TaggingError): pass
class TagFileFormatError(TaggingError): pass



# audiofile.py - kantag audio file metadata read and write operations.
# Copyright (C) 2018 David Gasaway
# https://bitbucket.org/dgasaway/kantag/

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
import sys
import os
import mutagen.id3
import mutagen.oggvorbis
import mutagen.flac
import mutagen.easymp4
import tagmaps, exceptions
from util import TagValue
from tagset import TagSet

# --------------------------------------------------------------------------------------------------
def _map_tag(tag, warn):
    """
    Get the cannonical name for a tag.
    """
    work = tag.lower()
    if work in tagmaps.general_read_map:
        work = tagmaps.general_read_map[work].lower()
    l = [t for t in tagmaps.cannonical_tags if t.lower() == work]
    if len(l) == 0:
        if warn:
            print >> sys.stderr, 'warning: unrecognized tag: ' + tag
        return tag
    else:
        return l[0]

# --------------------------------------------------------------------------------------------------
def _break_text_frame(frame, tag):
    """
    Break a mutagen text frame into a list of TagValue named tuples.
    """
    result = []
    if isinstance(frame.text, list):
        for value in frame.text:
            if isinstance(value, mutagen.id3.ID3TimeStamp):
                result.append(TagValue(tag, value.text))
            else:
                result.append(TagValue(tag, value))
    else:
        result.append(TagValue(tag, value))

    return result

# --------------------------------------------------------------------------------------------------
def _break_tipl_frame(frame, warn):
    """
    Break a mutagen TIPL ID3 frame into a list of TagValue named tuples.
    """
    # TIPL has a list of people/involvements.
    result = []
    for involvement, artist in frame.people:
        # A couple of TIPL involvements need a map to get to the cannonical tag, the rest we hope
        # are in the cannonical list.
        tag = involvement.lower()
        if tag in tagmaps.tipl_map:
            tag = tagmaps.tipl_map[tag]
        else:
            tag = _map_tag(tag, warn)
        result.append(TagValue(tag, artist))

    return result

# --------------------------------------------------------------------------------------------------
def _break_tmcl_frame(frame, tag):
    """
    Break a mutagen TMCL ID3 frame into a list of TagValue named tuples.
    """
    # TMCL has a list of people/instruments.
    result = []
    for instrument, artist in frame.people:
        # kantag format can't currently directly support instruments, so we'll map this to a
        # musicbrainz/vcomment credit where the data is: "Performer=Artist Name (instrument)".
        if instrument is not None and instrument != '':
            artist = artist + ' (' + instrument + ')'
        result.append(TagValue(tag, artist))

    return result

# --------------------------------------------------------------------------------------------------
def _break_ufid_frame(frame, tag):
    """
    Break a mutagen UFID ID3 frame into a list of TagValue named tuples.
    """
    return [TagValue(tag, frame.data)]

# --------------------------------------------------------------------------------------------------
def _break_apic_frame(frame, tag):
    """
    Break a mutagen APIC ID3 frame into a list of TagValue named tuples.
    """
    return [TagValue(tag, repr(frame.type))]

# --------------------------------------------------------------------------------------------------
def _break_frame(frame, ftype, warn):
    """
    Break a mutagen ID3 frame into a list of TagValue named tuples.
    """
    # COMM can have different attributes like COMM:description:eng and so on, but we'll drop all
    # that since it can't be reproduced in a tag file.
    if ftype.startswith('COMM:'):
        ftype = 'COMM'
    # Similar with APIC.
    if ftype.startswith('APIC:'):
        ftype = 'APIC'

    if ftype in tagmaps.id3_read_map:
        tag = tagmaps.id3_read_map[ftype]

        # Examine the type of frame to pull out the values correctly, and flatten where necessary.
        if isinstance(frame, mutagen.id3.TextFrame):
            return _break_text_frame(frame, tag)
        elif isinstance(frame, mutagen.id3.TIPL):
            return _break_tipl_frame(frame, warn)
        elif isinstance(frame, mutagen.id3.TMCL):
            return _break_tmcl_frame(frame, tag)
        elif isinstance(frame, mutagen.id3.UFID):
            return _break_ufid_frame(frame, tag)
        elif isinstance(frame, mutagen.id3.APIC):
            return _break_apic_frame(frame, tag)
        else:
            raise exceptions.TaggingError(
                'unexpected frame object encountered: ' + repr(type(frame))
                )
    else:
        if warn:
            print >> sys.stderr, 'warning: unknown ID3 frame type: ' + ftype
        return [TagValue(ftype, repr(frame))]

# --------------------------------------------------------------------------------------------------
def _build_frame(tag, values):
    """
    Build an mutagen ID3 frame from a tag name and associated values.
    """
    if tag in tagmaps.id3_write_map:
        frame = tagmaps.id3_write_map[tag]
    elif tag == 'TCMP':
        return mutagen.id3.TCMP(encoding=3, text=values)
    else:
        frame = 'TXXX:' + tag.upper()

    if frame == 'TALB':
        return mutagen.id3.TALB(encoding=3, text=values)
    elif frame == 'TPE1':
        return mutagen.id3.TPE1(encoding=3, text=values)
    elif frame == 'TCOM':
        return mutagen.id3.TCOM(encoding=3, text=values)
    elif frame == 'TPE3':
        return mutagen.id3.TPE3(encoding=3, text=values)
    elif frame == 'TDRC':
        return mutagen.id3.TDRC(encoding=3, text=values)
    elif frame == 'TPOS':
        return mutagen.id3.TPOS(encoding=3, text=values)
    elif frame == 'TSST':
        return mutagen.id3.TSST(encoding=3, text=values)
    elif frame == 'TCON':
        return mutagen.id3.TCON(encoding=3, text=values)
    elif frame == 'TEXT':
        return mutagen.id3.TEXT(encoding=3, text=values)
    elif frame == 'TIT2':
        return mutagen.id3.TIT2(encoding=3, text=values)
    elif frame == 'TRCK':
        return mutagen.id3.TRCK(encoding=3, text=values)
    elif frame == 'TSOP':
        return mutagen.id3.TSOP(encoding=3, text=values)
    elif frame == 'TIT3':
        return mutagen.id3.TIT3(encoding=3, text=values)
    elif frame == 'TSOC':
        return mutagen.id3.TSOC(encoding=3, text=values)
    elif frame == 'TDOR':
        return mutagen.id3.TDOR(encoding=3, text=values)
    elif frame.startswith('TXXX:'):
        desc = frame.partition(':')[2]
        return mutagen.id3.TXXX(encoding=3, desc=desc, text=values)
    elif frame.startswith('UFID:'):
        owner = frame.partition(':')[2]
        if len(values) > 1:
            raise exceptions.TaggingError('UFID owner not unique: ' + frame)
        return mutagen.id3.UFID(encoding=3, owner=owner, data=values[0])
    elif frame.startswith('COMM:'):
        split = frame.split(':', 3)
        return mutagen.id3.COMM(encoding=3, desc=split[1], lang=split[2], text=values)
    else:
        raise exceptions.TaggingError('unexpected mapped frame encountered: ' + frame)

    return None

# --------------------------------------------------------------------------------------------------
def _read_ogg(path, warn):
    """
    Read the existing tags from an ogg file, and return a list of TagValue named tuples.
    """
    result = []
    afile = mutagen.oggvorbis.OggVorbis(path)
    for item in afile.tags:
        tag = _map_tag(item[0], warn)
        result.append(TagValue(tag, '<BINARY DATA>' if tag == u'EmbeddedImage' else item[1]))

    return result
    #return [TagValue(_map_tag(v[0], warn), v[1]) for v in afile.tags]

# --------------------------------------------------------------------------------------------------
def _read_flac(path, warn):
    """
    Read the existing tags from a flac file, and return a list of TagValue named tuples.
    """
    # Ordinarily, FLAC stored embedded images in a separate block form the tags.  However, since
    # they do use vcomment tags, we'll check for the same as used in ogg vorbis.
    result = []
    afile = mutagen.flac.FLAC(path)
    for item in afile.tags:
        tag = _map_tag(item[0], warn)
        result.append(TagValue(tag, '<BINARY DATA>' if tag == u'EmbeddedImage' else item[1]))

    return result
    #return [TagValue(_map_tag(v[0], warn), v[1]) for v in afile.tags]

# --------------------------------------------------------------------------------------------------
def _read_mp3(path, warn):
    """
    Read the existing tags from an mp3 file, and return a list of TagValue named tuples.
    """
    result = []
    afile = mutagen.id3.ID3(path)
    afile.update_to_v24()
    for ftype, frame in afile.iteritems():
        result.extend(_break_frame(frame, ftype, warn))

    return result

# --------------------------------------------------------------------------------------------------
def _read_m4a(path, warn):
    """
    Read the existing tags from an m4a file, and return a list of TagValue named tuples.
    """
    # Register additional keys not supported by EasyMP4 by default.
    for name, key in tagmaps.mp4_map.items():
        mutagen.easymp4.EasyMP4Tags.RegisterFreeformKey(key, name)

    # Note, embedded images are not stored in tags.
    result = []
    afile = mutagen.easymp4.EasyMP4(path)
    for key, values in afile.iteritems():
        result.extend([TagValue(_map_tag(key, warn), v) for v in values])
    return result

# --------------------------------------------------------------------------------------------------
def read_raw(path, warn=True):
    """
    Read the existing tags from an audio file, and return a list of TagValue named tuples.
    """
    ext = os.path.splitext(path)[1].lower()
    if ext == '.ogg':
        return _read_ogg(path, warn)
    elif ext == '.flac':
        return _read_flac(path, warn)
    elif ext == '.mp3':
        return _read_mp3(path, warn)
    elif ext == '.m4a':
        return _read_m4a(path, warn)
    else:
        raise exceptions.FileTypeError('invalid file extension: ' + ext)

# --------------------------------------------------------------------------------------------------
def read(path, warn=True):
    """
    Read the existing tags from an audio file, and return a TagSet.
    """
    return TagSet(read_raw(path, warn))

# --------------------------------------------------------------------------------------------------
def _write_ogg(path, tagset):
    """
    Write tags from a TagSet to an ogg file.
    """
    afile = mutagen.oggvorbis.OggVorbis(path)
    afile.clear()
    for tag, values in tagset.iteritems():
        afile[tag.lower()] = values
    afile.save()

# --------------------------------------------------------------------------------------------------
def _write_flac(path, tagset):
    """
    Write tags from a TagSet to a flac file.
    """
    afile = mutagen.flac.FLAC(path)
    afile.clear()
    for tag, values in tagset.iteritems():
        afile[tag.lower()] = values
    afile.save()

# --------------------------------------------------------------------------------------------------
def _write_mp3(path, tagset):
    """
    Write tags from a TagSet to an mp3 file.
    """
    try:
        afile = mutagen.id3.ID3(path)
    except mutagen.id3.ID3NoHeaderError:
        afile = mutagen.id3.ID3()

    afile.clear()
    for tag, values in tagset.iteritems():
        frame = _build_frame(tag, values)
        if frame is not None:
            afile.add(frame)

    # Assume compilation and write a TCMP frame if an albumartist is present.
    #if u'AlbumArtist' in tags:
    #    file.add(_build_frame('TCMP', '1'))

    afile.save(path)

# --------------------------------------------------------------------------------------------------
def _write_m4a(path, tagset):
    """
    Write tags from a TagSet to an m4a file.
    """
    # Register additional keys not supported by EasyMP4 by default.
    for name, key in tagmaps.mp4_map.items():
        mutagen.easymp4.EasyMP4Tags.RegisterFreeformKey(key, name)

    afile = mutagen.easymp4.EasyMP4(path)
    afile.delete()  # Needed to remove tags not mapped by EasyMP4.
    afile.clear()
    for tag, values in tagset.iteritems():
        afile[tag.lower()] = values
    afile.save()

# --------------------------------------------------------------------------------------------------
def write(path, tagset):
    """
    Write tags from a TagSet to an audio file.
    """
    ext = os.path.splitext(path)[1].lower()
    if ext == '.ogg':
        _write_ogg(path, tagset)
    elif ext == '.flac':
        _write_flac(path, tagset)
    elif ext == '.mp3':
        _write_mp3(path, tagset)
    elif ext == '.m4a':
        _write_m4a(path, tagset)
    else:
        raise exceptions.FileTypeError('invalid file extension: ' + ext)



# tagfile.py - kantag text metadata file parser and builder.
# Copyright (C) 2018 David Gasaway
# https://bitbucket.org/dgasaway/kantag/

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
import sys
import re
import exceptions
import util
from listdict import ListDict
from tagset import TagSet
from tagstores import Release, Disc, Track
from tagmaps import cannonical_tags

# --------------------------------------------------------------------------------------------------
class TagLine(object):
    """
    Represents a line from a kantag tag file.
    """
    def __init__(self, line=None, line_type=None, applies_to=None, tag=None, value=None, warn=True):
        self._warn = warn

        # Initialize everything before attempting a parse.
        self._source_line = line
        self._line_type = None
        self._applies_to = None
        self._tag = None
        self._value = None
        self._used = False

        # If given a line, parse it.
        if self._source_line is not None:
            self._source_line = self._source_line.strip()
            if len(self._source_line) > 0:
                self._parse(self._source_line)

        # Let the remainder or parameters override what may have been parsed.
        if line_type is not None:
            self.line_type = line_type
        if applies_to is not None:
            self._applies_to = applies_to
        if tag is not None:
            self.tag = tag
        if value is not None:
            self._value = value

    # ----------------------------------------------------------------------------------------------
    @property
    def source_line(self):
        """Raw text source line passed to constructor."""
        return self._source_line

    # ----------------------------------------------------------------------------------------------
    @property
    def line_type(self):
        """Line type '#', 'a', 'd', or 't'."""
        return self._line_type
    @line_type.setter
    def line_type(self, value):
        self._line_type = value.lower()

    # ----------------------------------------------------------------------------------------------
    @property
    def applies_to(self):
        """List of numeric values the line applies to."""
        return self._applies_to
    @applies_to.setter
    def applies_to(self, value):
        self._applies_to = value

    # ----------------------------------------------------------------------------------------------
    @property
    def tag(self):
        """Tag name."""
        return self._tag
    @tag.setter
    def tag(self, value):
        self._tag = value
        if self.warn and value not in cannonical_tags:
            print >> sys.stderr, 'warning: unrecognized tag: ' + value.encode('utf-8')

    # ----------------------------------------------------------------------------------------------
    @property
    def value(self):
        """Tag value."""
        return self._value
    @value.setter
    def value(self, newvalue):
        self._value = newvalue

    # ----------------------------------------------------------------------------------------------
    @property
    def used(self):
        """Whether a tag line is used in the process of writing tags to audio files."""
        return self._used
    @used.setter
    def used(self, value):
        self._used = value

    # ----------------------------------------------------------------------------------------------
    @property
    def warn(self):
        """Whether warnings are enabled."""
        return self._warn
    @warn.setter
    def warn(self, value):
        self._warn = value

    # ----------------------------------------------------------------------------------------------
    def _parse(self, line):
        """
        Parse a kantag line into a TagLine.
        """
        if line.startswith('#'):
            # Comment line.
            pattern = r'^(?P<type>[#])(?P<comment>.*)$'
            match = re.match(pattern, line)
            if not match:
                raise exceptions.TagFileFormatError(
                    'Malformed comment tag: ' + line.encode('utf-8'))

            self.line_type = match.group('type')
            self._value = match.group('comment').strip()
            self._used = True

        elif line.startswith('a'):
            # Album/release line.
            pattern = r'^(?P<type>[aA]) (?P<tagname>[^=]*)=(?P<tagvalue>.+)$'
            match = re.match(pattern, line)
            if not match:
                raise exceptions.TagFileFormatError(
                    'Malformed album tag: ' + line.encode('utf-8'))

            self.line_type = match.group('type')
            self.tag = match.group('tagname')
            self._value = match.group('tagvalue')

        else:
            # Disc or track line.
            pattern = r'^(?P<type>[dtDT]) ' + \
                r'(?P<range>(\d+(-\d+)?)(,(\d+(-\d+)?))*) ' + \
                r'(?P<tagname>[^=]*)=(?P<tagvalue>.+)$'
            match = re.match(pattern, line)
            if not match:
                raise exceptions.TagFileFormatError(
                    'Malformed disc/track tag: ' + line.encode('utf-8'))

            self.line_type = match.group('type')
            self._applies_to = util.expand_ranges(match.group('range'))
            self.tag = match.group('tagname')
            self._value = match.group('tagvalue')

    # ----------------------------------------------------------------------------------------------
    def __unicode__(self):
        """
        Build a unicode kantag line from the current state.
        """
        if self._line_type is None:
            return u''
        elif self._line_type == '#':
            return u'# %s' % (self._value)
        elif self._line_type == 'a':
            return u'a %s=%s' % (self._tag, self._value)
        elif self._line_type == 'd' or self._line_type == 't':
            return u'%s %s %s=%s' % (self._line_type, self._applies_to, self._tag, self._value)
        else:
            raise exceptions.TagFileFormatError('Unexpected line type: ' + self._line_type)

    # ----------------------------------------------------------------------------------------------
    def __str__(self):
        """
        Build a UTF-8 encoded kantag line from the current state.
        """
        return unicode(self).encode('utf-8')

    # ----------------------------------------------------------------------------------------------
    def pprint(self):
        s = self._line_type + ': ' + \
            (', '.join(self._applies_to) if self._applies_to is not None else '[n/a]') + '\t' + \
            (self._tag + '=' if self._tag is not None else '') + \
            (self._value if self._value is not None else '[no value]')
        return s

# --------------------------------------------------------------------------------------------------
class TagFile(object):
    """
    Represents a collection of lines from a kantag tag file.
    """
    def __init__(self, warn=True):
        self._lines = []

    # ----------------------------------------------------------------------------------------------
    @property
    def lines(self):
        """List of contained TagLines."""
        return self._lines
    @lines.setter
    def lines(self, value):
        self._lines = value

    # ----------------------------------------------------------------------------------------------
    def apply_map(self, tag_map):
        """
        Apply a tag map to the lines of the file.
        """
        remove = []
        for line in self._lines:
            if line.tag in tag_map:
                new_tag = tag_map[line.tag]
                if new_tag is None:
                    remove.append(line)
                else:
                    line.tag = new_tag
        for line in remove:
            self._lines.remove(line)

    # ----------------------------------------------------------------------------------------------
    def get_matching(self, disc, track):
        """
        Get a TagSet of tags that should be applied to a track with a given disc number and track
        number.  If a line is used in the process, the 'used' attribue is set to True.
        """
        # Note: This approach of looping through the lines in one pass has the advantage of
        # preserving the order of the values for a given tag name as they appear in the file.
        disctrack = (track if disc is None else disc + track)
        result = TagSet()
        for line in self._lines:
            if line.line_type == 'a' or \
                line.line_type == 'd' and disc in line.applies_to or \
                line.line_type == 't' and disctrack in line.applies_to:
                    # Record the tag pair.
                    result.append(line.tag, line.value)
                    # Flag the line as used.
                    line.used = True

        return result

    # ----------------------------------------------------------------------------------------------
    def __unicode__(self):
        """
        Build unicode kantag lines from the current state.
        """
        return ('\n').join([unicode(l) for l in self._lines])

    # ----------------------------------------------------------------------------------------------
    def __str__(self):
        """
        Build UTF-8 encoded kantag lines from the current state.
        """
        return unicode(self).encode('utf-8')

    # ----------------------------------------------------------------------------------------------
    def pprint(self, print_comments=False):
        s = ''
        for line in [l for l in self._lines if print_comments or l.line_type != '#']:
            s += line.pprint() + '\n'
        return s

# --------------------------------------------------------------------------------------------------
class TagFileBuilder(object):
    """
    Helper for building a TagFile.
    """
    def __init__(self, tag_file=None, reader=None, warn=True):
        self._warn = warn
        self._file = tag_file
        if self._file is None:
            self._file = TagFile(warn=warn)
        if reader is not None:
            self._file.lines = \
                [TagLine(line, warn=warn) for line in reader.readlines() if len(line.strip()) > 0]

    # ----------------------------------------------------------------------------------------------
    @property
    def tags(self):
        """The underlying TagFile object."""
        return self._file

    # ----------------------------------------------------------------------------------------------
    @property
    def warn(self):
        """Whether warnings are enabled."""
        return self._warn
    @warn.setter
    def warn(self, value):
        self._warn = value

    # ----------------------------------------------------------------------------------------------
    def _get_value_entity_dict(self, entities, key):
        """
        Builds a dictionary keyed by the values of 'key' in all 'entities', where the value at each
        key is a list of entities with that value.  Result is a ListDict instance.
        """
        values = ListDict()
        for entity in entities:
            if key in entity.tags:
                for value in entity.tags[key]:
                    values.append_unique(value, entity)
        return values
    
    # ----------------------------------------------------------------------------------------------
    def _get_number_str(self, entities, parent=None):
        """
        Builds a number identifier for a TagStore entity or list of TagStore entities.  If a
        'entities' is a list, all items should be of the same type.
        """
        if not isinstance(entities, list):
            entities = [entities]
        
        nums = []
        for entity in entities:
            if isinstance(entity, Release):
                pass
            elif isinstance(entity, Disc):
                nums.append('%s' % entity.number)
            elif isinstance(entity, Track):
                if parent is not None and parent.number is not None:
                    nums.append('%s%s' % (parent.number, entity.number.zfill(2)))
                else:
                    nums.append('%s' % entity.number.zfill(2))
            else:
                raise exceptions.TaggingError('Unexpected entity type: ' + str(type(entity)))

        # Condense consecutive numbers into ranges.  Note that values are strings with leading
        # zeros.
        return util.condense_ranges(nums)

    # ----------------------------------------------------------------------------------------------
    def _add_value(self, entity, num, tag, value):
        """
        Add a TagLine built from a TagStore entity, a number identifier, a tag, and a value.
        """
        if isinstance(entity, Release):
            l = TagLine(None, 'a', num, tag, value, self.warn)
        elif isinstance(entity, Disc):
            l = TagLine(None, 'd', num, tag, value, self.warn)
        elif isinstance(entity, Track):
            l = TagLine(None, 't', num, tag, value, self.warn)
        else:
            raise exceptions.TaggingError('Unexpected entity type: ' + str(type(entity)))
        self._file.lines.append(l)

    # ----------------------------------------------------------------------------------------------
    def add_value(self, entity, tag, value, parent=None):
        """
        Add a TagLine built from a TagStore entity or list of TagStore entities, a tag, and a value.
        If 'entity' is a list, all items should be of the same type.
        """
        num = self._get_number_str(entity, parent)
        self._add_value(entity[0] if isinstance(entity, list) else entity, num, tag, value)

    # ----------------------------------------------------------------------------------------------
    def add_values(self, entity, key, parent=None):
        """
        Add TagLines built from tag values for a given key stored in a TagStore entity or list of
        TagStore entities.  If 'entity' is a list, all items should be of the same type.
        """
        if isinstance(entity, list):
            # First, build a dictionary that has the entities for each value.
            ved = self._get_value_entity_dict(entity, key);
            # Then sort them by the condensed number list.
            sorted_values = sorted(ved, key=lambda value: self._get_number_str(ved[value], parent))
            # Now output them in order.
            for value in sorted_values:
                self.add_value(ved[value], key, value, parent)
        elif key in entity.tags:
            for value in entity.tags[key]:
                self.add_value(entity, key, value, parent)

    # ----------------------------------------------------------------------------------------------
    def add_values_as(self, entity, key, as_tag, parent=None):
        """
        Add a TagLine built from a TagStore entity and its stored tag values for the key, but as an
        alternate tag name.
        """
        if key in entity.tags:
            for value in entity.tags[key]:
                self.add_value(entity, as_tag, value, parent)

    # ----------------------------------------------------------------------------------------------
    def add_values_req(self, entity, key, parent=None):
        """
        Add a TagLine built from a TagStore entity and its stored tag values for the key; a line is
        added even if no tags are present for the key.
        """
        if key in entity.tags:
            for value in entity.tags[key]:
                self.add_value(entity, key, value, parent)
        else:
            self.add_value(entity, key, '', parent)

    # ----------------------------------------------------------------------------------------------
    def add_values_as_req(self, entity, key, as_tag, parent=None):
        """
        Add a TagLine built from a TagStore entity and its stored tag values for the key, but as an
        alternate tag name; a line is added even if no tags are present for the key.
        """
        if key in entity.tags:
            for value in entity.tags[key]:
                self.add_value(entity, as_tag, value, parent)
        else:
            self.add_value(entity, as_tag, '', parent)

    # ----------------------------------------------------------------------------------------------
    def add_comment(self, comment):
        """
        Add a comment TagLine.
        """
        self._file.lines.append(TagLine(line_type='#', value=comment, warn=self.warn))

    # ----------------------------------------------------------------------------------------------
    def add_blank(self):
        """
        Add an empty TagLine, but only if the file is non-empty and the previous line is not blank.
        """
        cnt = len(self._file.lines)
        #print 'last: ' + self._file.lines[cnt-1].line_type
        if cnt > 0 and not self._file.lines[cnt-1].line_type is None:
            self._file.lines.append(TagLine(warn=self.warn))



# tagmaps.py - kantag metadata mapping defintions.
# Copyright (C) 2018 David Gasaway
# https://bitbucket.org/dgasaway/kantag/

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.

# TMCL='Musician credits list' (2.4)
# IPLS='Involved people list' (2.4)
# Had both TMCL and IPLS in Beastie Boys - Sure Shot ?!!!?
# TDRC replaced TDAT (2.4), TDAT could only do 4 chars.
# TDRC 'when the audio was recorded'
# TDRL 'when the audio was first released'
# Foobar: "The TDRC frame is officially to be used for the recording time, TDRL
#          is for the release time but TDRC is generally used by most programs."
# http://wiki.musicbrainz.org/MusicBrainz_Picard/Tags/Mapping
# http://musicbrainz.org/doc/Picard_Tag_Mapping
# http://musicbrainz.org/doc/MusicBrainzTag
# http://wiki.hydrogenaudio.org/index.php?title=Foobar2000:ID3_Tag_Mapping
# http://wiki.hydrogenaudio.org/index.php?title=Foobar2000:Encouraged_Tag_Standards#ALBUM_ARTIST
# http://wiki.hydrogenaudio.org/index.php?title=Foobar2000:Title_Formatting_Reference#Field_remappings
# http://code.google.com/p/mutagen/source/browse/trunk/mutagen/easyid3.py
# http://www.id3.org/id3v2.4.0-frames
# http://www.id3.org/id3v2.4.0-structure
# http://forums.musicbrainz.org/viewtopic.php?id=2359
# http://www.hydrogenaudio.org/forums/lofiversion/index.php/t85165.html

"""
Set of internal/kantag file tag names.
"""
cannonical_tags = frozenset([
    u'Album',
    u'AlbumArtist',
    u'AlbumArtistNonSort',
    u'AlbumArtistSort',
    u'AlbumArtists',
    u'AlbumArtistsSort',
    u'Arranger',
    u'ArrangerSort',
    u'Artist',
    u'ArtistNonSort',
    u'ArtistSort',
    u'Artists',
    u'ArtistsSort',
    u'ASIN',
    u'Barcode',
    u'CatalogNumber',
    u'Comment',
    u'Compilation',
    u'Composer',
    u'ComposerSort',
    u'Conductor',
    u'ConductorSort',
    u'CoverNumber',
    u'Date',
    u'DiscNumber',
    u'DiscSubtitle',
    u'DJMixer',
    u'EmbeddedImage',
    u'Engineer',
    u'Genre',
    u'ISRC',
    u'Label',
    u'LabelId',
    u'Language',
    u'Lyricist',
    u'LyricistSort',
    u'Media',
    u'Mixer',
    u'OriginalDate',
    u'OriginalYear',
    u'Part',
    u'Performer',
    u'PerformerSort',
    u'Producer',
    u'ReleaseCountry',
    u'ReleaseStatus',
    u'ReleaseType',
    u'Script',
    u'Title',
    u'TotalDiscs',
    u'TotalTracks',
    u'TrackNumber',
    u'Version',
    u'Work',
    u'Writer',
    u'WriterSort',
    u'acoustid_id',
    u'musicbrainz_albumartistid',
    u'musicbrainz_albumid',
    u'musicbrainz_albumstatus',
    u'musicbrainz_albumtype',
    u'musicbrainz_artistid',
    u'musicbrainz_discid',
    u'musicbrainz_releasegroupid',
    u'musicbrainz_trackid',
    u'musicbrainz_releasetrackid',
    u'musicbrainz_workid',
    u'replaygain_album_peak',
    u'replaygain_album_gain',
    u'replaygain_track_peak',
    u'replaygain_track_gain'
    ])

""" Map ID3 frames to kantag name. """
id3_read_map = {
    u'TALB': u'Album',
    u'TXXX:ALBUM ARTIST': u'AlbumArtist',
    u'TXXX:ALBUMARTIST': u'AlbumArtist',
    u'TPE2': u'AlbumArtist',
    u'TXXX:ALBUMARTISTNONSORT': u'AlbumArtistNonSort',
    u'TXXX:ALBUMARTISTSORT': u'AlbumArtistSort',
    u'TSO2': u'AlbumArtistSort',
    u'TXXX:ALBUMARTISTS': u'AlbumArtists',
    u'TXXX:ALBUMARTISTSSORT': u'AlbumArtistsSort',
    u'TXXX:albumartists': u'AlbumArtists',
    u'TXXX:albumartistssort': u'AlbumArtistsSort',
    u'TXXX:AMAZON_ID': u'AmazonId',
    u'TXXX:ARRANGER': u'Arranger',
    u'TXXX:ARRANGERSORT': u'ArrangerSort',
    u'TPE4': u'Arranger',
    u'TPE1': u'Artist',
    u'TXXX:ARTISTSORT': u'ArtistSort',
    u'TXXX:ARTISTNONSORT': u'ArtistNonSort',
    u'TSOP': u'ArtistSort',
    u'TXXX:Artists': u'Artists',
    u'TXXX:COMMENT': u'Comment',
    u'COMM': u'Comment',
    u'TCOM': u'Composer',
    u'TSOC': u'ComposerSort',
    u'TXXX:COMPOSERSORT': u'ComposerSort',
    u'TPE3': u'Conductor',
    u'TXXX:CONDUCTORSORT': u'ConductorSort',
    u'TXXX:COVERNUMBER': u'CoverNumber',
    u'TDRC': u'Date',
    u'TXXX:DATE': u'Date',
    u'TPOS': u'DiscNumber',
    u'TSST': u'DiscSubtitle',
    u'TCON': u'Genre',
    u'TPUB': u'Label',
    u'TXXX:LABELID': u'LabelId',
    u'TXXX:BARCODE': u'Barcode',
    u'TXXX:CATALOGNUMBER': u'CatalogNumber',
    u'TXXX:ASIN': u'LabelId',
    u'TLAN': u'Language',
    u'TEXT': u'Lyricist',
    u'TOLY': u'Lyricist',
    u'TXXX:LYRICISTSORT': u'LyricistSort',
    u'TMED': u'Media',
    u'TDOR': u'OriginalDate',
    u'TXXX:originalyear': u'OriginalDate',
    u'TXXX:PART': u'Part',
    u'TXXX:PERFORMER': u'Performer',
    u'TMCL': u'Performer',
    u'TXXX:PERFORMERSORT': u'PerformerSort',
    u'TXXX:MusicBrainz Album Release Country': u'ReleaseCountry',
    u'TXXX:MusicBrainz Album Status': u'ReleaseStatus',
    u'TXXX:MusicBrainz Album Type': u'ReleaseType',
    u'TXXX:SCRIPT': u'Script',
    u'TIT2': u'Title',
    u'TRCK': u'TrackNumber',
    u'TIT3': u'Version',
    u'TXXX:VERSION': u'Version',
    u'TXXX:version': u'Version',
    u'TXXX:WORK': u'Work',
    u'TXXX:Work': u'Work',
    u'TOAL': u'Work',
    u'TXXX:WRITER': u'Writer',
    u'TXXX:WRITERSORT': u'WriterSort',
    u'TXXX:writer': u'Writer',
    u'TXXX:writersort': u'WriterSort',
    u'TXXX:Writer': u'Writer',
    u'TCMP': u'Compilation',
    u'TIPL': u'Person', # See TIPL map below.
    u'TXXX:MusicBrainz Album Id': u'musicbrainz_albumid',
    u'TXXX:MusicBrainz Album Artist Id': u'musicbrainz_albumartistid',
    u'TXXX:MusicBrainz Artist Id': u'musicbrainz_artistid',
    u'TXXX:MusicBrainz Disc Id': u'musicbrainz_discid',
    u'TXXX:MusicBrainz Release Group Id': u'musicbrainz_releasegroupid',
    u'TXXX:MusicBrainz Release Track Id': u'musicbrainz_releasetrackid',
    u'UFID:http://musicbrainz.org': u'musicbrainz_trackid',
    u'TXXX:MusicBrainz Work Id': u'musicbrainz_workid',
    u'TXXX:MUSICBRAINZ_TRACKID': u'musicbrainz_trackid',
    u'TXXX:replaygain_album_peak': u'replaygain_album_peak',
    u'TXXX:replaygain_album_gain': u'replaygain_album_gain',
    u'TXXX:replaygain_track_peak': u'replaygain_track_peak',
    u'TXXX:replaygain_track_gain': u'replaygain_track_gain',
    u'TXXX:Acoustid Id': u'acoustid_id',
    u'APIC': u'EmbeddedImage'
    }

""" Map non-ID3 tags from, e.g., Picard to cannonical names. """
# Note that the destination value should be something in cannonical_tags.
general_read_map = {
    u'remixer': u'Arranger',
    u'djmixer': u'Arranger',
    u'disctotal': u'TotalDiscs',
    u'tracktotal': u'TotalTracks',
    u'metadata_block_picture': u'EmbeddedImage'
    }

""" Map TIPL involvements to kantag name. """
tipl_map = {
    u'dj-mix': u'DJMixer',
    u'mix': u'Mixer'
    }

""" Map kantag name to ID3 frame. """
id3_write_map = {
    u'Album': u'TALB',
    u'AlbumArtist': u'TXXX:ALBUM ARTIST',
    u'Artist': u'TPE1',
    u'ArtistSort': u'TSOP',
    # Having more than one COMM with the same content descriptor (third descriptor, after
    #     language) can cause problems with, for example, foobar2000. "There may be more
    #     than one comment frame in each tag, but only one with the same language and
    #     content descriptor."
    # 'Comment': 'COMM::eng', # having more than one with the same content
    u'Comment': u'TXXX:COMMENT',
    u'Composer': u'TCOM',
    u'Conductor': u'TPE3',
    u'Date': u'TDRC',
    u'DiscNumber': u'TPOS',
    u'DiscSubtitle': u'TSST',
    u'Genre': u'TCON',
    u'Language': u'TLAN',
    u'Lyricist': u'TEXT',
    u'OriginalDate': u'TDOR',
    u'Title': u'TIT2',
    u'TrackNumber': u'TRCK',
    u'musicbrainz_albumid': u'TXXX:MusicBrainz Album Id',
    u'musicbrainz_albumartistid': u'TXXX:MusicBrainz Album Artist Id',
    u'musicbrainz_artistid': u'TXXX:MusicBrainz Artist Id',
    u'musicbrainz_releasegroupid': u'TXXX:MusicBrainz Release Group Id',
    u'musicbrainz_releasetrackid': u'TXXX:MusicBrainz Release Track Id',
    u'musicbrainz_trackid': u'UFID:http://musicbrainz.org',
    u'musicbrainz_workid': u'TXXX: MusicBrainz Work Id',
    u'replaygain_album_peak': u'TXXX:replaygain_album_peak',
    u'replaygain_album_gain': u'TXXX:replaygain_album_gain',
    u'replaygain_track_peak': u'TXXX:replaygain_track_peak',
    u'replaygain_track_gain': u'TXXX:replaygain_track_gain'
    # Anything not explicitly listed maps to TXXX:<WHATEVER>
    }

""" Additional freeform items for writing to MP4/M4A. """
mp4_map = {
    # Note: 'albumartistsort' is supported by EasyMP4, but with default settings foobar2000 will
    #       map 'soaa' as ALBUMARTISTSORTORDER.
    b'albumartistsort': 'albumartistsort',
    b'albumartists': 'albumartists',
    b'albumartistssort': 'albumartistssort',
    b'arranger': 'arranger',
    b'arrangersort': 'arrangersort',
    b'composer': 'composer',
    # Note: 'composersort' is supported by EasyMP4, but with default settings foobar2000 will map
    #       'soco' as COMPOSERSORTORDER.
    b'composersort': 'composersort',
    b'conductor': 'conductor',
    b'conductorsort': 'conductorsort',
    b'covernumber': 'covernumber',
    b'discsubtitle': 'discsubtitle',
    b'labelid': 'labelid',
    b'lyricist': 'lyricist',
    b'lyricistsort': 'lyricistsort',
    b'originaldate': 'originaldate',
    b'part': 'part',
    b'performer': 'performer',
    b'performersort': 'performersort',
    b'version': 'version',
    b'work': 'work',
    b'writer': 'writer',
    b'writersort': 'writersort',
    b'replaygain_album_peak': 'replaygain_album_peak',
    b'replaygain_album_gain': 'replaygain_album_gain',
    b'replaygain_track_peak': 'replaygain_track_peak',
    b'replaygain_track_gain': 'replaygain_track_gain'
}

""" vcomment/ID3 map used by foobar2000; for development reference only. """
_foobar_id3_map = {
    'album': 'TALB',
    'albumartist': 'TXXX:ALBUM ARTIST',
    'artist': 'TPE1',
    'artistsortorder': 'TSOP', # This causes problems for mixed vcomment/mp3 fb2k library
    'discnumber': 'TPOS',
    'band': 'TPE2',
    'comment': 'COMM',
    'composer': 'TCOM',
    'conductor': 'TPE3',
    'date': 'TDRC', # also, TXXX:DATE
    'discnumber': 'TPOS',
    'genre': 'TCON',
    'tracknumber': 'TRCK',
    'remixed by': 'TPE4',
    'lyricist': 'TEXT',
    'subtitle': 'TIT3',
    'replaygain_album_peak': 'TXXX:replaygain_album_peak',
    'replaygain_album_gain': 'TXXX:replaygain_album_gain',
    'replaygain_track_peak': 'TXXX:replaygain_track_peak',
    'replaygain_track_gain': 'TXXX:replaygain_track_gain',
    # Anything not explicitly listed maps to TXXX:<WHATEVER>
    }

""" Tag remaps used by foobar2000; for development reference only. """
_foobar_field_remap = (
    ['album artist', ['album artist', 'artist', 'composer', 'performer', 'albumartist']],
    ['album', ['album', 'venue']],
    ['artist', ['artist', 'album artist', 'composer', 'performer']],
    ['discnumber', ['discnumber', 'disc']],
    ['title', ['title', 'filename']]
    )

# NOTE: comment, discnumber, and discsubtitle are not set by stock Picard.
""" vcomment/ID3 map used by Musicbrainz Picard; for development reference only. """
_picard_id3_map = {
    'album': 'TALB',
    'albumartist': 'TPE2',
    'albumartistsort': 'TXXX:ALBUMARTISTSORT',
    'arranger': 'TIPL:arranger',
    'artist': 'TPE1',
    'artistsort': 'TSOP',
    'asin': 'TXXX:ASIN',
    'barcode': 'TXXX:BARCODE',
    'catalognumber': 'TXXX:CATALOGNUMBER',
    'composer': 'TCOM',
    'comment': 'COMM:description',
    'compilation': 'TCMP',
    'conductor': 'TPE3',
    'date': 'TDRC',
    'discnumber': 'TPOS',
    'discsubtitle': 'TSST',
    'djmixer': 'TIPL:DJ-mix',
    'engineer': 'TIPL:engineer',
    'label': 'TPUB',
    'language': 'TLAN',
    'lyricist': 'TEXT',
    'mixer': 'TIPL:mix',
    'originaldate': 'TDOR',
    'performer': 'TMCL:instrument',
    'producer': 'TIPL:producer',
    'remixer': 'TPE4',
    'subtitle': 'TIT3',
    'title': 'TIT2',
    'tracknumber': 'TRCK',
    'writer': 'TXXX:writer',
    'musicbrainz_trackid': 'UFID:http://musicbrainz.org',
    'musicbrainz_albumid': 'TXXX:MusicBrainz Album Id',
    'musicbrainz_artistid': 'TXXX:MusicBrainz Artist Id',
    'musicbrainz_albumartistid': 'TXXX:MusicBrainz Album Artist Id',
    'musicbrainz_workid': 'TXXX:MusicBrainz Work Id'
    }

""" vcomment/ID3 map used by mutagen; for development reference only. """
_mutagen_easyid3_map = {
    'genre': 'TCON',
    'date': 'TDRC',
    'performer:role': 'TMCL:role',
    'album': 'TALB',
    'composer': 'TCOM',
    'lyricist': 'TEXT',
    'title': 'TIT2',
    'version': 'TIT3',
    'artist': 'TPE1',
    'performer': 'TPE2',
    'conductor': 'TPE3',
    'arranger': 'TPE4',
    'discnumber': 'TPOS',
    'tracknumber': 'TRCK',
    'composersort': 'TSOC', # iTunes extension
    'artistsort': 'TSOP',
    'discsubtitle': 'TSST',
    'albumartistsort': 'TXXX:ALBUMARTISTSORT', # also TSO2, iTunes extension
    'barcode': 'TXXX:BARCODE',
    'musicbrainz_trackid': 'UFID:http://musicbrainz.org',
    'musicbrainz_albumid': 'TXXX:MusicBrainz Album Id',
    'musicbrainz_artistid': 'TXXX:MusicBrainz Artist Id',
    'musicbrainz_albumartistid': 'TXXX:MusicBrainz Album Artist Id',
    'musicbrainz_discid': 'TXXX:MusicBrainz Disc Id'
    }



# tagstores.py - kantag heirachical metadata containers and builders.
# Copyright (C) 2018 David Gasaway
# https://bitbucket.org/dgasaway/kantag/

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
import re
import sys
import os.path
import warnings
import pprint
from tagset import TagSet
import audiofile, util, textencoding
try:
    import musicbrainz as mb
except ImportError:
    mb = None

# --------------------------------------------------------------------------------------------------
class TagStore(object):
    """
    Base class that exposes an internal TagSet as a tags property.
    """
    def __init__(self):
        self._tags = TagSet()
        self._children = []

    # ----------------------------------------------------------------------------------------------
    @property
    def tags(self):
        """A TagSet of tag-values that apply to the entity."""
        return self._tags
    @tags.setter
    def tags(self, value):
        self._tags = value

# --------------------------------------------------------------------------------------------------
class Track(TagStore):
    """
    Container for track information and tags.
    """
    def __init__(self, number=None):
        if number is not None:
            self._tags[u'TrackNumber'] = [number]

    # ----------------------------------------------------------------------------------------------
    @property
    def number(self):
        """Track number."""
        return self._tags[u'TrackNumber'][0] if u'TrackNumber' in self._tags else None
    @number.setter
    def number(self, value):
        self._tags[u'TrackNumber'] = [value]

    # ----------------------------------------------------------------------------------------------
    def pprint(self):
        return 'TRACK: ' + (self.number if self.number is not None else 'None') + ':\n' + \
            pprint.PrettyPrinter(indent=2).pformat(self._tags) + '\n'

# --------------------------------------------------------------------------------------------------
class Disc(TagStore):
    """
    Container for disc information, tracks, and tags.
    """
    def __init__(self, number):
        TagStore.__init__(self)
        if number is not None:
            self._tags[u'DiscNumber'] = [number]
        self._is_single_artist = None
        self._is_various = None

    # ----------------------------------------------------------------------------------------------
    @property
    def number(self):
        """Disc number."""
        return self._tags[u'DiscNumber'][0] if u'DiscNumber' in self._tags else None
    @number.setter
    def number(self, value):
        self._tags[u'DiscNumber'] = [value]

    # ----------------------------------------------------------------------------------------------
    @property
    def is_single_artist(self):
        """
        True if the disc has just a single track artist.  Note that a release can have multiple
        track artists without being a Various Artist disc.
        """
        return self._is_single_artist
    @is_single_artist.setter
    def is_single_artist(self, value):
        self._is_single_artist = value

    # ----------------------------------------------------------------------------------------------
    @property
    def is_various(self):
        """True if the disc is Various Artist disc."""
        return self._is_various
    @is_various.setter
    def is_various(self, value):
        self._is_various = value

    # ----------------------------------------------------------------------------------------------
    @property
    def tracks(self):
        """List of Tracks on the disc."""
        return self._children
    @tracks.setter
    def tracks(self, value):
        self._children = value

    # ----------------------------------------------------------------------------------------------
    def replace_all(self, key, replace, replacement):
        """
        Replace a tag values in the disc and all children with a given key/value with a
        replacement value.
        """
        self.tags.replace(key, replace, replacement)
        for track in self.tracks:
            track.tags.replace(key, replace, replacement)

    # ----------------------------------------------------------------------------------------------
    def pprint(self):
        s = 'DISC: ' + (self.number if self.number is not None else 'None') + \
            ' [' + ('Single-artist' if self.is_single_artist else 'Multi-artist') + \
            ('/Various' if self.is_various else '') + ']' + '\n' + \
            pprint.PrettyPrinter(indent=2).pformat(self.tags) + '\n'
        for track in self.tracks:
            s += track.pprint()
        return s

# --------------------------------------------------------------------------------------------------
class Release(TagStore):
    """
    Container for release information, discs, and tags.
    """
    def __init__(self):
        TagStore.__init__(self)
        self._is_single_artist = None
        self._is_various = None

    # ----------------------------------------------------------------------------------------------
    @property
    def is_single_artist(self):
        """
        True if the release has just a single track artist.  Note that a release can have multiple
        track artists without being a Various Artist release.
        """
        return self._is_single_artist
    @is_single_artist.setter
    def is_single_artist(self, value):
        self._is_single_artist = value

    # ----------------------------------------------------------------------------------------------
    @property
    def is_various(self):
        """True if the release is Various Artist release."""
        return self._is_various
    @is_various.setter
    def is_various(self, value):
        self._is_various = value

    # ----------------------------------------------------------------------------------------------
    @property
    def discs(self):
        """List of Discs contained in the release."""
        return self._children
    @discs.setter
    def discs(self, value):
        self._children = value

    # ----------------------------------------------------------------------------------------------
    def replace_all(self, key, replace, replacement):
        """
        Replace a tag values in the release and all children with a given key/value with a
        replacement value.
        """
        self.tags.replace(key, replace, replacement)
        for disc in self.discs:
            disc.replace_all(key, replace, replacement)

    # ----------------------------------------------------------------------------------------------
    def pprint(self):
        s = 'RELEASE [' + ('Single-artist' if self.is_single_artist else 'Multi-artist') + \
            ('/Various' if self.is_various else '') + \
            ('/Multi-disc' if len(self.discs) > 1 else '') + ']' + '\n' + \
            pprint.PrettyPrinter(indent=2).pformat(self.tags) + '\n'
        for disc in self.discs:
            s += disc.pprint() + '\n'
        return s

# --------------------------------------------------------------------------------------------------
class _TagStoreBuilder(object):
    """
    Base class for TagStore build helpers.
    """
    def __init__(self, options, entity=None):
        self._options = options
        self._entity = entity

    # ----------------------------------------------------------------------------------------------
    def apply_musicbrainz_relation(self, relation):
        """
        Apply a musicbrainz Relation to the given TagSet.
        """
        # If the tag already contains this sortname for this particular arist, we'll replace it with
        # the regular name, and vice versa (useful for working with some legacy data).
        entity = self._entity
        entity.tags.append_replace(relation.type, relation.sortname, relation.name)
        entity.tags.append_replace(relation.type + u'Sort', relation.name, relation.sortname)

    # ----------------------------------------------------------------------------------------------
    def apply_musicbrainz_relations(self, relations):
        """
        Apply a list musicbrainz Relation instances to the given TagSet.
        """
        map(self.apply_musicbrainz_relation, relations)

    # ----------------------------------------------------------------------------------------------
    def _set_artist_statuses(self, common_values, various):
        """
        Set is_single_artist and is_various flags based on the common values produced in an entity
        merge operation.  If it's determined that this is a VA entity, then AlbumArtist and
        AlbumArtistSort tags may get added to the common values.
        """
        entity = self._entity
        if entity.is_single_artist is None:
            entity.is_single_artist = u'Artist' in common_values
        entity.is_various = \
            not entity.is_single_artist and \
            (u'AlbumArtist' not in common_values or \
            u'Various Artists' in common_values[u'AlbumArtist'] or \
            u'Various' in common_values[u'AlbumArtist'] or \
            various in common_values[u'AlbumArtist'])
        if entity.is_various and not u'AlbumArtist' in common_values:
            entity.tags.append(u'AlbumArtist', various)
            entity.tags.append(u'AlbumArtistSort', various)

    # ----------------------------------------------------------------------------------------------
    def _merge_children(self, various, keep_common):
        """
        Add tag/value pairs that are common to all children to the entity.
        """
        # First get the common values and add them to the entity tags.
        entity = self._entity
        common_values = TagSet.get_common_values([child.tags for child in entity._children])
        entity.tags.merge_unique(common_values)

        # If the single-artist/various statuses couldn't be determined by a earlier edits, then look
        # look for a common artist tag, and set the property values for convenience.
        self._set_artist_statuses(common_values, various)

        # Remove the common child values from the originating children.
        for child in entity._children:
            child.tags.remove_dict(common_values)

# --------------------------------------------------------------------------------------------------
class TrackBuilder(_TagStoreBuilder):
    """
    Helper for building a Track object from a file.
    """
    def __init__(self, options, track=None):
        _TagStoreBuilder.__init__(self, options, track)
        if self._entity is None:
            self._entity = Track()

    # ----------------------------------------------------------------------------------------------
    @property
    def track(self):
        """The underlying Track object."""
        return self._entity

    # ----------------------------------------------------------------------------------------------
    def read(self, filename):
        """
        Read tags from file.  This will overwrite any existing tags.
        """
        self.track.tags = audiofile.read(filename)

    # ----------------------------------------------------------------------------------------------
    def infer_from_path(self, filename, regex):
        """
        Use file path to add missing artist, album, date, discnumber, tracknumber, title tags.  If
        the regular expression fails to match, the filename less extension is used as the title.
        """
        path = os.path.abspath(filename)
        match = re.search(regex, path)
        tags = self.track.tags
        if match:
            parts = match.groupdict()
            if u'Artist' not in tags and 'artist' in parts and parts['artist'] != '':
                tags[u'Artist'] = [parts['artist']]
            if u'Album' not in tags and 'album' in parts and parts['album'] != '':
                tags[u'Album'] = [parts['album']]
            if u'Date' not in tags and 'date' in parts and parts['date'] != '':
                tags[u'Date'] = [parts['date']]
            if u'DiscNumber' not in tags and 'disc' in parts and parts['disc'] != '':
                tags[u'DiscNumber'] = [parts['disc']]
            if u'TrackNumber' not in tags and 'track' in parts and parts['track'] != '':
                tags[u'TrackNumber'] = [parts['track'].zfill(2)]
            # Also want to see if the existing title is a generic value like "Track 5".
            if (u'Title' not in tags or tags[u'Title'][0][0:6] == 'Track ') and \
                'title' in parts and parts['title'] !='':
                tags[u'Title'] = [parts['title']]
        else:
            if u'Title' not in tags:
                tags[u'Title'] = [os.path.splitext(os.path.basename(path))[0]]

    # ----------------------------------------------------------------------------------------------
    def split_disc_title(self):
        """
        Parse a musicbrainz-style disc title into title and disc number/subtitle.
        """
        tags = self.track.tags
        if u'Album' in tags:
            for title in tags[u'Album']:
                parsed = util.parse_album_title(unicode(title))
                if parsed.discnum is not None:
                    tags[u'Album'] = [parsed.title]
                    tags[u'DiscNumber'] = [parsed.discnum]
                if parsed.subtitle is not None:
                    tags[u'DiscSubtitle'] = [parsed.subtitle]

    # ----------------------------------------------------------------------------------------------
    def split_title_to_work_and_parts(self, classical):
        """
        Parse a musicbrainz-style title into work and part.
        """
        tags = self.track.tags
        if u'Title' in tags:
            for title in tags[u'Title']:
                parsed = util.parse_track_title(unicode(title))
                if parsed.parts is not None:
                    tags.remove(u'Title', title)
                    if parsed.work is not None:
                        #tags.append_unique(u'Work', parsed.work)
                        tags[u'Work'] = [parsed.work]
                    for part in parsed.parts:
                        tags.append_unique(u'Part', part)
                if u'Work' not in tags and classical:
                    tags.move_values(u'Title', u'Work')

    # ----------------------------------------------------------------------------------------------
    def remove_instruments(self):
        """
        Remove instrument/role information from Performer tags.
        """
        tags = self.track.tags
        if u'Performer' in tags:
            tags[u'Performer'] = util.remove_artist_roles(tags[u'Performer'])

    # ----------------------------------------------------------------------------------------------
    def _apply_release_data(self, mb_release):
        """
        Apply release-level musicbrainz metadata to track tags.
        """
        tags = self.track.tags

        artists = mb.get_artists(mb_release, self._options.locale)
        tags[u'AlbumArtists'] = [artist.name for artist in artists]
        tags[u'AlbumArtistsSort'] = [artist.sortname for artist in artists]
        if len(artists) == 1:
            tags[u'AlbumArtist'] = [artists[0].name]
            tags[u'AlbumArtistSort'] = [artists[0].sortname]

        tags[u'Album'] = [unicode(mb_release['title'])]
        
        # Use the earliest (US) release date and earliest release group release date.
        date = mb.get_earliest_release(mb_release)
        if date is not None:
            tags[u'Date'] = [date]
        orig_date = mb.get_earliest_rg_release(mb_release)
        if orig_date is not None:
            if u'Date' in tags:
                if orig_date != tags[u'Date'][0]:
                    tags[u'OriginalDate'] = [orig_date]
            else:
                tags[u'Date'] = [date]
        
        # Barcode, ASIN, CatalogNumber.
        if 'asin' in mb_release:
            tags[u'ASIN'] = [mb_release['asin']]
        if 'barcode' in mb_release and mb_release['barcode'] != '':
            tags[u'Barcode'] = [mb_release['barcode']]
        catnum = mb.get_release_catalognumber(mb_release)
        if not catnum is None:
            tags[u'CatalogNumber'] = [catnum]
        
        # Add the artist relations using sortnames.
        self.apply_musicbrainz_relations(mb.get_artist_relations(mb_release, self._options.locale))
        
    # ----------------------------------------------------------------------------------------------
    def _apply_medium_data(self, mb_medium):
        """
        Apply medium-level musicbrainz metadata to track tags.
        """
        tags = self.track.tags
        if 'title' in mb_medium:
            tags[u'DiscSubtitle'] = [unicode(mb_medium['title'])]

    # ----------------------------------------------------------------------------------------------
    def _apply_track_data(self, mb_track):
        """
        Apply track-level musicbrainz metadata to track tags.  Returns whether a track title was
        found.
        """
        tags = self.track.tags

        found_title = ('title' in mb_track)
        if found_title:
            tags[u'Title'] = [unicode(mb_track['title'])]

        artists = mb.get_artists(mb_track, self._options.locale)
        tags[u'Artists'] = [artist.name for artist in artists]
        tags[u'ArtistsSort'] = [artist.sortname for artist in artists]
        if len(artists) == 1:
            tags[u'Artist'] = [artists[0].name]
            tags[u'ArtistSort'] = [artists[0].sortname]
        
        return found_title

    # ----------------------------------------------------------------------------------------------
    def _apply_recording_data(self, mb_recording, apply_title):
        """
        Apply recording-level musicbrainz metadata to the track tags.  'apply_title' indicates
        whether recording title should overwrite an existing title tag.
        """
        tags = self.track.tags

        tags[u'RecordingTitle'] = [unicode(mb_recording['title'])]
        if (not u'Title' in tags or apply_title) and 'title' in mb_recording:
            tags[u'Title'] = [unicode(mb_recording['title'])]
        if 'disambiguation' in mb_recording and mb_recording['disambiguation'] != '':
            tags.append_unique(u'Version', mb_recording['disambiguation'])

        # Add the artist relations using sortnames.
        locale = self._options.locale
        self.apply_musicbrainz_relations(mb.get_artist_relations(mb_recording, locale))

    # ----------------------------------------------------------------------------------------------
    def _apply_work_data(self, mb_work, recurse, apply_title):
        tags = self.track.tags

        #mb_work = mb.get_work_by_id(mb_work['id']);
        if apply_title:
            title = unicode(mb_work['title'])
            if self._options.parse_title:
                parsed = util.parse_track_title(title)
                if parsed.parts is not None:
                    if parsed.work is not None:
                        tags.append_unique(u'Work', parsed.work)
                        #tags[u'Work'] = [parsed.work]
                    for part in parsed.parts:
                        tags.append_unique(u'Part', part)
            else:
                tags[u'Work'] = tags.append_unique(u'Work', title)


        # Add the artist relations using sortnames.
        rels = mb.get_work_artist_relations(mb_work, recurse, self._options.locale)
        #pprint.pprint(rels, stream=sys.stderr)
        self.apply_musicbrainz_relations(rels)

    # ----------------------------------------------------------------------------------------------
    def apply_musicbrainz(self, mbdata):
        """
        Apply musicbrainz metadata to the track tags.  'mbdata' should be a release instance
        returned by the musicbrainz API.  The track needs 'TrackNumber' and 'DiscNumber' tags to
        receive medium, track, recording, and work metadata.
        """
        # Bail out of musicbrainz is not available.
        if mb is None:
            warnings.warn('musicbrainz package is not available')
        
        # Note: All data is initially loaded into track tags - from existing tags, inferred from
        # path, etc.  On the other hand, musicbrainz data is hierarchical.  Which means, in order
        # to edit existing data with musicbrainz data, we need to either first merge track data to
        # "unflatten" before applying musicbrainz data, or first "flatten" the musicbrainz data to
        # track level.  Earlier revisions used the former approach, but this didn't always work out
        # as desired, and was more difficult to implement.  Now, we use the latter approach, and
        # rely on the built-in "merging" logic rebuild the heirarchy post-edit.  Thus, musicbrainz
        # data is applied to track objects only.

        tags = self.track.tags
        
        self._apply_release_data(mbdata)
        medium_num = (tags[u'DiscNumber'][0] if u'DiscNumber' in tags else '1')
        mb_medium = mb.get_release_medium(mbdata, medium_num)
        if not mb_medium is None:
            self._apply_medium_data(mb_medium)

            mb_track = mb.get_medium_track(mb_medium, self.track.number)
            if not mb_track is None:
                # If the track title matches the recording title, then the API will not include a
                # title on the track, so we need to pull it from the recording instead.
                found_title = self._apply_track_data(mb_track)

                mb_recording = mb.get_track_recording(mb_track)
                if not mb_recording is None:
                    self._apply_recording_data(mb_recording, not found_title)

                    # Only apply the work title from the first.
                    apply_title = True
                    for mb_work in mb.get_recording_works(mb_recording):
                        # TODO: Make work recursion an option.
                        self._apply_work_data(mb_work, False, apply_title)
                        apply_title = False

# --------------------------------------------------------------------------------------------------
class DiscBuilder(_TagStoreBuilder):
    """
    Helper for building a Disc from files.
    """
    def __init__(self, options, disc=None):
        _TagStoreBuilder.__init__(self, options, disc)
        if self._entity is None:
            self._entity = Disc()

    # ----------------------------------------------------------------------------------------------
    @property
    def disc(self):
        """The underlying Disc object."""
        return self._entity

    # ----------------------------------------------------------------------------------------------
    def merge_tracks(self, various, keep_common):
        """
        Add tag/value pairs that are common to all tracks to the disc.
        """
        self._merge_children(various, keep_common)

# --------------------------------------------------------------------------------------------------
class ReleaseBuilder(_TagStoreBuilder):
    """
    Helper for building a Release heirarchy from files.
    """
    def __init__(self, options, release=None):
        _TagStoreBuilder.__init__(self, options, release)
        self._musicbrainz_data = None
        if self._entity is None:
            self._entity = Release()

    # ----------------------------------------------------------------------------------------------
    @property
    def release(self):
        """The underlying Release object."""
        return self._entity

    # ----------------------------------------------------------------------------------------------
    @property
    def musicbrainz_data(self):
        """MusicBrainz API data about the release."""
        return self._musicbrainz_data
    @musicbrainz_data.setter
    def musicbrainz_data(self, value):
        self._musicbrainz_data = value

    # ----------------------------------------------------------------------------------------------
    def merge_discs(self):
        """
        Add tag/value pairs that are common to all discs to the release.
        """
        self._merge_children(self._options.various, self._options.keep_common)

    # ----------------------------------------------------------------------------------------------
    def _get_track(self, filename):
        """
        Get the tags for file sourced from the existing tags in the file, the filename and path (if
        the existing tags have gaps and can be inferred from filename/path).  Also does some release
        and track title parsing based on musicbrainz-originated titles.
        """
        builder = TrackBuilder(self._options)
        builder.read(filename)

        # Fill in gaps using filename/path.
        if self._options.infer_from_path:
            builder.infer_from_path(filename, self._options.path_regex)

        # Remove roles from artist names.
        if self._options.remove_instruments:
            builder.remove_instruments()

        # Parse the album title in case it is a musicbrainz-style multi-disc title.
        if self._options.parse_disc:
            builder.split_disc_title()

        # Temporarily make sure everything is 'Various Artists' so we don't have issues with unique
        # appends or replaces with mismatched values.
        tags = builder.track.tags
        tags.replace(u'AlbumArtist', u'Various', u'Various Artists')
        tags.replace(u'AlbumArtist', self._options.various, u'Various Artists')
        tags.replace(u'AlbumArtistSort', u'Various', u'Various Artists')
        tags.replace(u'AlbumArtistSort', self._options.various, u'Various Artists')
        tags.replace(u'AlbumArtists', u'Various', u'Various Artists')
        tags.replace(u'AlbumArtists', self._options.various, u'Various Artists')
        tags.replace(u'AlbumArtistsSort', u'Various', u'Various Artists')
        tags.replace(u'AlbumArtistsSort', self._options.various, u'Various Artists')

        # Note: Some old musicbrainz metadata will have a different albumid for each disc in a
        # multi-disc release.  This is an artifact of the days when musicbrainz represented each
        # disc as a different release.  However, the API will return the same full release metadata
        # for either mbid.
        if self._options.call_musicbrainz:
            mb._set_api_format(self._options.api_format)
            if self.musicbrainz_data is None and u'musicbrainz_albumid' in tags:
                self.musicbrainz_data = mb.get_release_by_id(tags[u'musicbrainz_albumid'][0])

        if not self.musicbrainz_data is None:
            builder.apply_musicbrainz(self.musicbrainz_data)

        # Parse track title into work/part.
        if self._options.parse_title:
            builder.split_title_to_work_and_parts(self._options.classical)

        # All track data is now loaded.  Apply optional removing of non-ASCII punctuation.
        if self._options.ascii_punctuation:
            tags.apply_to_all('Album', textencoding.asciipunct)
            tags.apply_to_all('DiscSubtitle', textencoding.asciipunct)
            tags.apply_to_all('Title', textencoding.asciipunct)
            tags.apply_to_all('RecordingTitle', textencoding.asciipunct)
            tags.apply_to_all('Work', textencoding.asciipunct)
            tags.apply_to_all('Part', textencoding.asciipunct)
            tags.apply_to_all('Version', textencoding.asciipunct)

        return builder.track

    # ----------------------------------------------------------------------------------------------
    def _get_disc(self, track):
        """
        Build and/or return a Disc matching the given Track.  If a new Disc is needed, it is added
        to the release.  The Track is added to the Disc.
        """
        release = self.release

        # Check if this is the first encounter with a certain disc number.
        discnum = (track.tags[u'DiscNumber'][0] if u'DiscNumber' in track.tags else None)
        if not discnum in [d.number for d in release.discs]:
            # Initialize a new disc object.
            disc = Disc(discnum)
            release.discs.append(disc)
        else:
            disc = [d for d in release.discs if d.number == discnum][0]

        return disc

    # ----------------------------------------------------------------------------------------------
    def read(self, filenames):
        """
        Return a Release object containing the tags from the given files in a heirarchy.  Tag
        sources include existing tags, path information, and musicbrainz API calls.  Many edits are
        applied to conform the data as closely as possible to certain standards.
        """
        release = self.release

        for filename in filenames:
            if self._options.verbose >= 1:
                print >> sys.stderr, filename

            # Fetch the tags that are stored in the file/path.
            track = self._get_track(filename)

            # Get a new or existing Disc matching the track.
            disc = self._get_disc(track)
            disc.tracks.append(track)

        for disc in release.discs:
            # Merge values that are common to all tracks in a disc into the disc tags.
            dbuilder = DiscBuilder(self._options, disc)
            dbuilder.merge_tracks(self._options.various, self._options.keep_common)

        # Merge values that are common to all discs in the release into the release tags.
        self.merge_discs()

        # Set the final "Various Artists"/"Various" values.
        release.replace_all(u'AlbumArtist', u'Various Artists', self._options.various)
        release.replace_all(u'AlbumArtistSort', u'Various Artists', self._options.various)
        release.replace_all(u'AlbumArtists', u'Various Artists', self._options.various)
        release.replace_all(u'AlbumArtistsSort', u'Various Artists', self._options.various)

        return release



# tagset.py - kantag metadata storage primitive.
# Copyright (C) 2018 David Gasaway
# https://bitbucket.org/dgasaway/kantag/

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
from listdict import ListDict

class TagSet(ListDict):
    """
    Represents a set of tag/value sets that could be applied to an audio file.  Each item in the
    ListDict is keyed by a tag name, and contains a list of values for that name.
    """

    # --------------------------------------------------------------------------------------------------
    def __init__(self, tags=None):
        """
        Initializes an instance from a list of TagValue named tuples.
        """
        if tags is not None:
            for tagvalue in tags:
                self.append_unique(tagvalue.tag, tagvalue.value)



# musicbrainz.py - kantag musicbrainz operations.
# Copyright (C) 2018 David Gasaway
# https://bitbucket.org/dgasaway/kantag/

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
import sys
import time
import re
import collections
import warnings
import musicbrainzngs as ngs
from kantag._version import __version__

# Globals
""" Maps musicbrainz relationship UUIDs to cannonical tags. """
_mbz_reltype_map = {
    # Work 'composer'
    'd59d99ea-23d4-4a80-b066-edca32ee158f' : u'Composer',
    # Work 'librettist'
    '7474ab81-486f-40b5-8685-3a4f8ea624cb' : u'Lyricist',
    # Work 'lyricist'
    '3e48faba-ec01-47fd-8e89-30e81161661c' : u'Lyricist',
    # Work 'translator'
    'da6c5d8a-ce13-474d-9375-61feb29039a5' : u'Lyricist',
    # Work 'arranger'
    'd3fd781c-5894-47e2-8c12-86cc0e2c8d08' : u'Arranger',
    # Work 'instrument arranger'
    '0084e70a-873e-4f7f-b3ff-635b9e863dae' : u'Arranger',
    # Work 'vocal arranger'
    '6a88b92b-8fb5-41b3-aa1f-169ee7e05ed6' : u'Arranger',
    # Work 'orchestrator'
    '0a1771e1-8639-4740-8a43-bdafc050c3da' : u'Arranger',
    # Work 'revised by'
    'eeb9c319-9fde-4313-b76d-29db1576aad8' : u'Arranger',
    # Work 'reconstructed by'
    'cb887d1b-5267-4f3d-badb-5b3fba7349f6' : u'Arranger',
    # Work 'writer'
    'a255bca1-b157-4518-9108-7b147dc3fc68' : u'Writer',
    
    # Recording 'arranger'
    '22661fb8-cdb7-4f67-8385-b2a8be6c9f0d' : u'Arranger',
    # Recording 'instrument arranger'
    '4820daa1-98d6-4f8b-aa4b-6895c5b79b27' : u'Arranger',
    # Recording 'vocal arranger'
    '8a2799e8-a7e2-41ce-a7da-b5f520687216' : u'Arranger',
    # Recording 'orchestrator'
    '38fa7405-f9a5-48cb-827a-8ac601933ba0' : u'Arranger',
    # Recording 'performer'
    '628a9658-f54c-4142-b0c0-95f031b544da' : u'Performer',
    # Recording 'instrument'
    '59054b12-01ac-43ee-a618-285fd397e461' : u'Performer',
    # Recording 'vocal'
    '0fdbe3c6-7700-4a31-ae54-b53f06ae1cfa' : u'Performer',
    # Recording 'performing orchestra'
    '3b6616c5-88ba-4341-b4ee-81ce1e6d7ebb' : u'Performer',
    # Recording 'programming'
    '36c50022-44e0-488d-994b-33f11d20301e' : u'Performer',
    # Recording 'conductor'
    '234670ce-5f22-4fd0-921b-ef1662695c5d' : u'Conductor',
    # Recording 'chorus master'
    '45115945-597e-4cb9-852f-4e6ba583fcc8' : u'Conductor',
    # Recording 'remixer'
    '7950be4d-13a3-48e7-906b-5af562e39544' : u'Arranger',    
    
    # Release 'composer'
    '01ce32b0-d873-4baa-8025-714b45c0c754' : u'Composer',
    # Release 'librettist'
    'dd182715-ca2b-4e4b-80b1-d21742fda0dc' : u'Lyricist',
    # Release 'lyricist'
    'a2af367a-b040-46f8-af21-310f92dfe97b' : u'Lyricist',
    # Release 'translator'
    '4db37fec-eb67-45d3-b4fa-148a68135fbb' : u'Lyricist',
    # Release 'arranger'
    '34d5334e-a4c8-4b65-a5f8-bbcc9c81d13d' : u'Arranger',
    # Release 'instrument arranger'
    '18f159bb-44f0-4aef-b198-a4736ad9b659' : u'Arranger',
    # Release 'vocal arranger'
    'd7d9128d-e676-4d8f-a353-f48a55a98501' : u'Arranger',
    # Release 'orchestrator'
    '04e1f0b6-ef40-4168-ba25-42a87729fe09' : u'Arranger',
    # Release 'writer'
    'ca7a474a-a1cd-4431-9230-56a17f553090' : u'Writer',
    # Release 'performer'
    '888a2320-52e4-4fe8-a8a0-7a4c8dfde167' : u'Performer',
    # Release 'instrument'
    '67555849-61e5-455b-96e3-29733f0115f5' : u'Performer',
    # Release 'vocal'
    'eb10f8a0-0f4c-4dce-aa47-87bcb2bc42f3' : u'Performer',
    # Release 'performing orchestra'
    '23a2e2e7-81ca-4865-8d05-2243848a77bf' : u'Performer',
    # Release 'programming'
    '617063ad-dbb5-4877-9ba0-ba2b9198d5a7' : u'Performer',
    # Release 'conductor'
    '9ae9e4d0-f26b-42fb-ab5c-1149a47cf83b' : u'Conductor',
    # Release 'chorus master'
    'b9129850-73ec-4af5-803c-1c12b97e25d2' : u'Conductor',
    # Release 'remixer'
    'ac6a86db-f757-4815-a07e-744428d2382b' : u'Arranger',    
    }
""" Musicbrainz relationship UUID for work-work 'parts' relationship. """
_mbz_parts_id = 'ca8d3642-ce5f-49f8-91f2-125d72524e6a'
""" Musicbrainz relationsip UUID for recording-work 'performance' relationship. """
_mbz_performance_id = 'a3005666-a872-32c3-ad06-98af558e99b0'

""" Representation of an artist relation. """
Relation = collections.namedtuple('Relation', 'type, name, sortname')
""" An artist name-sortname pair. """
ArtistName = collections.namedtuple('ArtistName', 'name, sortname')

""" API response format. """
_format = 'json'

# Initialize the user agent.
ngs.set_useragent('kantag', __version__, 'https://bitbucket.org/dgasaway/kantag/')

# --------------------------------------------------------------------------------------------------
def _set_api_format(api_format):
    """
    Set the API format (xml or json).  FOR DEBUG USE ONLY.
    """
    global _format
    _format = api_format
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        ngs.set_format(api_format)

# --------------------------------------------------------------------------------------------------
def get_release_by_id(releaseid):
    """
    Call the musicbrainz API and return the Release object matching the releaseid.
    """
    incs = ['artists', 'artist-credits', 'artist-rels', 'recordings', 'recording-level-rels',
            'work-rels', 'work-level-rels', 'release-groups', 'labels', 'aliases']
    result = ngs.get_release_by_id(releaseid, includes=incs)
    if _format == 'json':
        return result
    else:
        return result['release']

# --------------------------------------------------------------------------------------------------
def get_recording_by_id(recordingid):
    """
    Call the musicbrainz API and return recording information, including artist and work ARs.
    """
    incs = ['artist-rels', 'work-rels', 'aliases']
    result = ngs.get_recording_by_id(recordingid, includes=incs)
    if _format == 'json':
        return result
    else:
        return result['recording']

# --------------------------------------------------------------------------------------------------
def get_work_by_id(workid, include_work_rels=False):
    """
    Call the musicbrainz API and return work information, including artist ARs.
    """
    incs = ['artist-rels', 'work-rels', 'aliases']
    result = ngs.get_work_by_id(workid, includes=incs)
    if _format == 'json':
        return result
    else:
        return result['work']

# --------------------------------------------------------------------------------------------------
def get_artist_primary_alias(artist, locale='en'):
    """
    Extract the primary locale alias from an artist returned by the musicbrainz API.  The result is
    an ArtistName named tuple with 'Artist' as the type.  If no primary locale alias is found, a
    tuple containing the original artist is returned.
    """
    key = 'aliases' if _format == 'json' else 'alias-list'
    primary_val = True if _format == 'json' else 'primary'
    alias_name_key = 'name' if _format == 'json' else 'alias'
    
    if key in artist:
        for alias in artist[key]:
            if 'primary' in alias and alias['primary'] == primary_val:
                if 'locale' in alias and alias['locale'] == locale:
                    return ArtistName(alias[alias_name_key], alias['sort-name'])
    return ArtistName(artist['name'], artist['sort-name'])

# --------------------------------------------------------------------------------------------------
def map_relation(relation, locale='en'):
    """
    Map a musicbrainz artist relation to a tag name.  The result is a Relation named tuple.
    """
    rel_type = relation['type-id']
    if rel_type in _mbz_reltype_map:
        artist = get_artist_primary_alias(relation['artist'], locale)
        return Relation(_mbz_reltype_map[rel_type], artist.name, artist.sortname)
    else:
        return None

# --------------------------------------------------------------------------------------------------
def map_relations(relations, locale='en'):
    """
    Map a musicbrainz artist relations to a tag names.  The result is a list of Relation named
    tuples. A relation type that has no map will not be in the list.
    """
    names = [map_relation(rel, locale) for rel in relations]
    return [name for name in names if not name is None]

# --------------------------------------------------------------------------------------------------
def get_artists(entity, locale='en'):
    """
    Extract the artists from an entity (e.g., release) returned by the musicbrainz API.  The result
    is a list of Relation named tupes with 'Artist' as the type.
    """
    result = []
    if 'artist-credit' in entity:
        for ac in entity['artist-credit']:
            if type(ac) == type(dict()) and 'artist' in ac:
                artist = get_artist_primary_alias(ac['artist'], locale)
                result.append(Relation(u'Artist', artist.name, artist.sortname))
    elif 'artist-credit-phrase' in entity:
        result.append(Relation(u'Artist', entity['artist-credit-phrase'], ''))

    return result

# --------------------------------------------------------------------------------------------------
def get_artist_relations(entity, locale='en'):
    """
    Extract the artist relations from an entity (e.g., recording or release) returned by the
    musicbrainz API.  The result is a list of Relation named tuples.  A relation type that has no
    map will be excluded, and duplicates of the same type and artist are removed.
    """
    key = 'relations' if _format == 'json' else 'artist-relation-list'
    if key in entity:
        # Map the relations to tag/name tuples.
        return list(set(map_relations(entity[key], locale))) 
    else:
        return []

# --------------------------------------------------------------------------------------------------
def get_work_partof_works(work):
    """
    Extract the works of which the a work returned by the musicbrainz API is a part.  The result is
    a list of work objects.  Only examines subworks in the loaded metadata, and does not query the
    API.
    """
    key = 'relations' if _format == 'json' else 'work-relation-list'
    result = []
    if key in work:
        for rel in work[key]:
            if rel['type-id'] == _mbz_parts_id:
                if 'direction' in rel and rel['direction'] == 'backward':
                    result.append(rel['work'])
    return result

# --------------------------------------------------------------------------------------------------
def _get_work_artist_relations(work, recurse=False, locale='en'):
    """
    Extract the artist relationships from a work returned by the musicbrainz API, optionally
    recursively following 'part of' relationships making further API calls, as necessary.
    """
    result = get_artist_relations(work, locale)
    if not recurse:
        return result

    # Recurse through part-of relation parents and add to list.
    parents = get_work_partof_works(work)
    for parent in parents:
        work = get_work_by_id(parent['id'], True)
        result = result + _get_work_artist_relations(work, True, locale)
    
    return result

# --------------------------------------------------------------------------------------------------
def get_work_artist_relations(work, recurse=False, locale='en'):
    """
    Extract the artist relationships from a work returned by the musicbrainz API, optionally
    recursively following 'part of' relationships making further API calls, as necessary.  The
    passed API work data must contain the work-artist relations; to recurse, it must also contain
    one level of work-work relations.  The result is a list of Relation named tuples.  A relation
    type that has no map will be excluded, and duplicates of the same type and artist are removed.
    """
    return list(set(_get_work_artist_relations(work, recurse, locale)))

# --------------------------------------------------------------------------------------------------
def get_earliest_release(release):
    """
    Extract the earliest US release date or earliest release date if no US date is available, from
    a release returned by the musicbrainz API.  The result is a date as a string or None if no date
    was found.  Only looks at release events on the specific release.
    """
    earliest = None
    earliest_US = None
    key = 'release-events' if _format == 'json' else 'release-event-list'
    if key in release:
        for event in release[key]:
            if 'date' in event:
                if earliest is None or event['date'] < earliest:
                    earliest = event['date']
                is_US = ('iso-3166-1-code-list' in event and 'US' in event['iso-3166-1-code-list'])
                if is_US and (earliest_US is None or event['date'] < earliest_US):
                    earliest_US = event['date']

    return earliest if earliest_US is None else earliest_US

# --------------------------------------------------------------------------------------------------
def get_earliest_rg_release(release):
    """
    Extract the earliest release date for the release group associated with a release returned by
    the musicbrainz API.  The result is a date as a string or None if no date was found.
    """
    if 'release_group' in release and 'first-release-date' in release['release_group']:
        return release['release_group']['first-release-date']
    else:
        return None

# --------------------------------------------------------------------------------------------------
def get_release_catalognumber(release):
    """
    Extract a label catalog number from a release returned by the musicbrainz API.  The result is a
    catalog number as a string or None if no catalog number was found.
    """
    key = 'label-info' if _format == 'json' else 'label-info-list'
    if key in release:
        for label in release[key]:
            if 'catalog-number' in label:
                return label['catalog-number']
    return None

# --------------------------------------------------------------------------------------------------
def get_release_medium(release, number):
    """
    Extract a medium by number from a release returned by the musicbrainz API.  The result is a
    musicbrainz API medium object.
    """
    key = 'media' if _format == 'json' else 'medium-list'
    if key in release:
        for medium in release[key]:
            if int(medium['position']) == int(number):
                return medium
    return None

# --------------------------------------------------------------------------------------------------
def get_medium_track(medium, number):
    """
    Extract a track by number from a medium returned by the musicbrainz API.  The result is a
    musicbrainz API track object.
    """
    key = 'tracks' if _format == 'json' else 'track-list'
    if 'tracks' in medium:
        for track in medium['tracks']:
            if int(track['position']) == int(number):
                return track
    return None

# --------------------------------------------------------------------------------------------------
def get_track_recording(track):
    """
    Extract a recording from a track returned by the musicbrainz API.  The result is a musicbrainz
    API recording object.
    """
    if 'recording' in track:
        return track['recording']
    else:
        return None

# --------------------------------------------------------------------------------------------------
def get_recording_works(recording):
    """
    Extract the works associated with a recording returned by the musicbrainz API.  The result is a
    list of musicbrainz API work objects.
    """
    result = []
    key = 'relations' if _format == 'json' else 'work-relation-list'
    if key in recording:
        rels = recording[key]
        return [r['work'] for r in rels if r['type-id'] == _mbz_performance_id]
    else:
        return []



# listdict.py - kantag dictionary of lists.
# Copyright (C) 2018 David Gasaway
# https://bitbucket.org/dgasaway/kantag/

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
class ListDict(dict):
    """
    Data container that extends dict with convenience methods for storing a list with each
    dictionary slot.
    """
    # ----------------------------------------------------------------------------------------------
    @staticmethod
    def _list_diff(a, b):
        """
        Return a list of values that are in list a but not in list b, preserving element order.
        """
        return [value for value in a if not value in b]

    # ----------------------------------------------------------------------------------------------
    @staticmethod
    def _list_int(a, b):
        """
        Return a list of values that are in both list and list b, preserving element order.
        """
        return [value for value in a if value in b]

    # ----------------------------------------------------------------------------------------------
    @staticmethod
    def _list_remove(lst, value):
        """
        Remove all occurences of value from the list in-line.
        """
        i = len(lst) - 1
        while i >= 0:
            if lst[i] == value:
                lst.pop(i)
            i -= 1

    # ----------------------------------------------------------------------------------------------
    @staticmethod
    def _list_replace(lst, replace, replacement):
        """
        Replace all values matching values in-line in a list.  Return whether any values were
        replaced.
        """
        replaced = False
        for i, v in enumerate(lst):
            if v == replace:
                lst[i] = replacement
                replaced = True
        return replaced

    # ----------------------------------------------------------------------------------------------
    @staticmethod
    def _list_replace_ci(lst, replace, replacement):
        """
        Replace all values case-insensitive-matching values in-line in a list.  Return whether any
        values were replaced.
        """
        replaced = False
        for i, v in enumerate(lst):
            if v.lower() == replace.lower():
                lst[i] = replacement
                replaced = True
        return replaced

    # ----------------------------------------------------------------------------------------------
    def append(self, key, value):
        """
        Append a value to the list at the key.
        """
        if key in self:
            self[key].append(value)
        else:
            # Create a new sequence at the key.
            self[key] = [value]

    # ----------------------------------------------------------------------------------------------
    def append_unique(self, key, value):
        """
        Append a value to the list at they key, provided the value is not already in the list.
        """
        if key in self:
            l = self[key]
            if value not in l:
                l.append(value)
        else:
            # Create a new sequence at the key.
            self[key] = [value]

    # ----------------------------------------------------------------------------------------------
    def extend(self, key, values):
        """
        Append values to the list at the key.
        """
        for value in values:
            self.append(key, value)

    # ----------------------------------------------------------------------------------------------
    def extend_unique(self, key, values):
        """
        Append values to the list at the key, provided the values are not already in the list.
        """
        for value in values:
            self.append_unique(key, value)

    # ----------------------------------------------------------------------------------------------
    def merge_unique(self, dct):
        """
        Add values from a dictionary that are not already in the object.
        """
        for key, values in dct.iteritems():
            self.extend_unique(key, values)

    # ----------------------------------------------------------------------------------------------
    def replace(self, key, replace, replacement):
        """
        Replace matching values in the list at the key.
        """
        if key in self:
            #self[key] = map((lambda v: replacement if v == replace else v), self[key])
            ListDict._list_replace(self[key], replace, replacement)

    # ----------------------------------------------------------------------------------------------
    def replace_ci(self, key, replace, replacement):
        """
        Replace case-insensitive-matching values in the list at the key.
        """
        if key in self:
            #self[key] = map((lambda v: replacement if v == replace else v), self[key])
            ListDict._list_replace_ci(self[key], replace, replacement)

    # ----------------------------------------------------------------------------------------------
    def remove(self, key, value):
        """
        Remove a value from the list at key; if the list is empty after removing the value, remove
        the key.
        """
        if key in self:
            l = [v for v in self[key] if v != value]
            if len(l) == 0:
                del self[key]
            else:
                self[key] = l

    # ----------------------------------------------------------------------------------------------
    def append_replace(self, key, replace, value):
        """
        Replace matching values in the list at the key with another value.  If neither the replaced
        value nor replacement value is in the list, then append the replacement value.
        """
        # Note we want this code to preserve the order of the items in the list.  That's why we use
        # a replace when possible rather than a simpler remove/append.
        if key in self:
            l = self[key]
            replaced = ListDict._list_replace(l, replace, value)
            # As an optimization, avoid attempting to append if there was a replace.
            if not replaced and value not in l:
                l.append(value)
        else:
            self[key] = [value]

    # ----------------------------------------------------------------------------------------------
    def remove_dict(self, remove):
        """
        Remove all the values from the dictionary that are present in a second dictionary.
        """
        for key, values in remove.iteritems():
            for value in values:
                self.remove(key, value)

    # ----------------------------------------------------------------------------------------------
    def move_values(self, a, b):
        """
        Provided there is a list with key a, replace the list at key b with the list at key a and
        remove key a.
        """
        if a in self:
            self[b] = self[a]
            del self[a]

    # ----------------------------------------------------------------------------------------------
    def contains(self, key, value):
        """
        Return whether the list at the given key contains the given value.
        """
        return key in self and value in self[key]

    # ----------------------------------------------------------------------------------------------
    def apply_to_all(self, key, func):
        """
        Apply a function to each value in the list at key.
        """
        if key in self:
            self[key] = [func(value) for value in self[key]]

    # ----------------------------------------------------------------------------------------------
    @staticmethod
    def get_common_values(dicts):
        """
        Return a new ListDict that contains the values common to all the ListDicts passed.
        """
        result = ListDict()

        for key, values in dicts[0].iteritems():
            common_values = values[:]
            for dict2 in dicts[1:]:
                if key in dict2:
                    common_values = ListDict._list_int(common_values, dict2[key])
                else:
                    common_values = []
            if len(common_values) > 0:
                result[key] = common_values

        return result



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
# Copyright (C) 2004 Robert Kaye
# Copyright (C) 2006 Lukáš Lalinský
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

# This modules provides functionality for simplifying unicode strings.

# The unicode character set (of over 1m codepoints and 24,000 characters) includes:
#   Normal ascii (latin) non-accented characters
#   Combined latin characters e.g. ae in normal usage
#   Compatibility combined latin characters (retained for compatibility with other character sets)
#     These can look very similar to normal characters and can be confusing for searches, sort orders etc.
#   Non-latin (e.g. japanese, greek, hebrew etc.) characters
# Both latin and non-latin characters can be accented. Accents can be either:
#   Provided by separate nonspacing_mark characters which are visually overlaid (visually 1 character is actually 2); or
#   Integrated accented characters (i.e. non-accented characters combined with a nonspace_mark into a single character)
# Again these can be confusing for searches, sort orders etc.
# Punctuation can also be confusing in unicode e.g. several types of single or double quote mark.

# For latin script:
#   Combined characters, accents and punctuation can be visually similar but look different to search engines,
#   sort orders etc. and the number of ways to use similar looking characters can (does) result in inconsistent
#   usage inside Music metadata.
#
#   Simplifying # the unicode character sets by many-to-one mappings can improve consistency and reduce confusion,
#   however sometimes the choice of specific characters can be a deliberate part of an album, song title or artist name
#   (and should not therefore be changed without careful thought) and occasionally the choice of characters can be
#   malicious (i.e. to defeat firewalls or spam filters or to appear to be something else).
#
#   Finally, given the size of the unicode character set, fonts are unlikely to display all characters,
#   making simplification a necessity.
#
#   Simplification may also be needed to make tags conform to ISO-8859-1 (extended ascii) or to make tags or filenames
#   into ascii, perhaps because the file system or player cannot support unicode.
#
# Non-latin scripts may also need to be converted to latin scripts through:
#   Translation (e.g. hebrew word for mother is translated to "mother"); or
#   Transliteration (e.g. the SOUND of the hebrew letter or word is spelt out in latin)
# These are non-trivial, and the software to do these is far from comprehensive.

# This module provides utility functions to enable simplification of latin and punctuation unicode:
#   1. simplify compatibility characters;
#   2. split combined characters;
#   3. remove accents (entirely or if not in ISO-8859-1 as applicable);
#   4. replace remaining non-ascii or non-ISO-8859-1 characters with a default character
# This module also provides an extension infrastructure to allow translation and / or transliteration plugins to be added.

import re
import unicodedata
import codecs
from functools import partial

#########################  LATIN SIMPLIFICATION ###########################
# The translation tables for punctuation and latin combined-characters are taken from
# http://unicode.org/repos/cldr/trunk/common/transforms/Latin-ASCII.xml
# Various bugs and mistakes in this have been ironed out during testing.


def _re_any(iterable):
    return re.compile('([' + ''.join(iterable) + '])', re.UNICODE)

_additional_compatibility = {
    u"\u0276": u"Œ",  # LATIN LETTER SMALL CAPITAL OE
    u"\u1D00": u"A",  # LATIN LETTER SMALL CAPITAL A
    u"\u1D01": u"Æ",  # LATIN LETTER SMALL CAPITAL AE
    u"\u1D04": u"C",  # LATIN LETTER SMALL CAPITAL C
    u"\u1D05": u"D",  # LATIN LETTER SMALL CAPITAL D
    u"\u1D07": u"E",  # LATIN LETTER SMALL CAPITAL E
    u"\u1D0A": u"J",  # LATIN LETTER SMALL CAPITAL J
    u"\u1D0B": u"K",  # LATIN LETTER SMALL CAPITAL K
    u"\u1D0D": u"M",  # LATIN LETTER SMALL CAPITAL M
    u"\u1D0F": u"O",  # LATIN LETTER SMALL CAPITAL O
    u"\u1D18": u"P",  # LATIN LETTER SMALL CAPITAL P
    u"\u1D1B": u"T",  # LATIN LETTER SMALL CAPITAL T
    u"\u1D1C": u"U",  # LATIN LETTER SMALL CAPITAL U
    u"\u1D20": u"V",  # LATIN LETTER SMALL CAPITAL V
    u"\u1D21": u"W",  # LATIN LETTER SMALL CAPITAL W
    u"\u1D22": u"Z",  # LATIN LETTER SMALL CAPITAL Z
    u"\u3007": u"0",  # IDEOGRAPHIC NUMBER ZERO
    u"\u00A0": u" ",  # NO-BREAK SPACE
    u"\u3000": u" ",  # IDEOGRAPHIC SPACE (from ‹character-fallback›)
    u"\u2033": u"”",  # DOUBLE PRIME
}
_re_additional_compatibility = _re_any(_additional_compatibility.keys())


def unicode_simplify_compatibility(string):
    interim = _re_additional_compatibility.sub(lambda m: _additional_compatibility[m.group(0)], string)
    return unicodedata.normalize("NFKC", interim)


_simplify_punctuation = {
    u"\u013F": u"L",  # LATIN CAPITAL LETTER L WITH MIDDLE DOT (compat)
    u"\u0140": u"l",  # LATIN SMALL LETTER L WITH MIDDLE DOT (compat)
    u"\u2018": u"'",  # LEFT SINGLE QUOTATION MARK (from ‹character-fallback›)
    u"\u2019": u"'",  # RIGHT SINGLE QUOTATION MARK (from ‹character-fallback›)
    u"\u201A": u"'",  # SINGLE LOW-9 QUOTATION MARK (from ‹character-fallback›)
    u"\u201B": u"'",  # SINGLE HIGH-REVERSED-9 QUOTATION MARK (from ‹character-fallback›)
    u"\u201C": u"\"",  # LEFT DOUBLE QUOTATION MARK (from ‹character-fallback›)
    u"\u201D": u"\"",  # RIGHT DOUBLE QUOTATION MARK (from ‹character-fallback›)
    u"\u201E": u"\"",  # DOUBLE LOW-9 QUOTATION MARK (from ‹character-fallback›)
    u"\u201F": u"\"",  # DOUBLE HIGH-REVERSED-9 QUOTATION MARK (from ‹character-fallback›)
    u"\u2032": u"'",  # PRIME
    u"\u2033": u"\"",  # DOUBLE PRIME
    u"\u301D": u"\"",  # REVERSED DOUBLE PRIME QUOTATION MARK
    u"\u301E": u"\"",  # DOUBLE PRIME QUOTATION MARK
    u"\u00AB": u"<<",  # LEFT-POINTING DOUBLE ANGLE QUOTATION MARK (from ‹character-fallback›)
    u"\u00BB": u">>",  # RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK (from ‹character-fallback›)
    u"\u2039": u"<",  # SINGLE LEFT-POINTING ANGLE QUOTATION MARK
    u"\u203A": u">",  # SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
    u"\u00AD": u"",  # SOFT HYPHEN (from ‹character-fallback›)
    u"\u2010": u"-",  # HYPHEN (from ‹character-fallback›)
    u"\u2011": u"-",  # NON-BREAKING HYPHEN (from ‹character-fallback›)
    u"\u2012": u"-",  # FIGURE DASH (from ‹character-fallback›)
    u"\u2013": u"-",  # EN DASH (from ‹character-fallback›)
    u"\u2014": u"-",  # EM DASH (from ‹character-fallback›)
    u"\u2015": u"-",  # HORIZONTAL BAR (from ‹character-fallback›)
    u"\uFE31": u"|",  # PRESENTATION FORM FOR VERTICAL EM DASH (compat)
    u"\uFE32": u"|",  # PRESENTATION FORM FOR VERTICAL EN DASH (compat)
    u"\uFE58": u"-",  # SMALL EM DASH (compat)
    u"\u2016": u"||",  # DOUBLE VERTICAL LINE
    u"\u2044": u"/",  # FRACTION SLASH (from ‹character-fallback›)
    u"\u2045": u"[",  # LEFT SQUARE BRACKET WITH QUILL
    u"\u2046": u"]",  # RIGHT SQUARE BRACKET WITH QUILL
    u"\u204E": u"*",  # LOW ASTERISK
    u"\u3008": u"<",  # LEFT ANGLE BRACKET
    u"\u3009": u">",  # RIGHT ANGLE BRACKET
    u"\u300A": u"<<",  # LEFT DOUBLE ANGLE BRACKET
    u"\u300B": u">>",  # RIGHT DOUBLE ANGLE BRACKET
    u"\u3014": u"[",  # LEFT TORTOISE SHELL BRACKET
    u"\u3015": u"]",  # RIGHT TORTOISE SHELL BRACKET
    u"\u3018": u"[",  # LEFT WHITE TORTOISE SHELL BRACKET
    u"\u3019": u"]",  # RIGHT WHITE TORTOISE SHELL BRACKET
    u"\u301A": u"[",  # LEFT WHITE SQUARE BRACKET
    u"\u301B": u"]",  # RIGHT WHITE SQUARE BRACKET
    u"\uFE11": u",",  # PRESENTATION FORM FOR VERTICAL IDEOGRAPHIC COMMA (compat)
    u"\uFE12": u".",  # PRESENTATION FORM FOR VERTICAL IDEOGRAPHIC FULL STOP (compat)
    u"\uFE39": u"[",  # PRESENTATION FORM FOR VERTICAL LEFT TORTOISE SHELL BRACKET (compat)
    u"\uFE3A": u"]",  # PRESENTATION FORM FOR VERTICAL RIGHT TORTOISE SHELL BRACKET (compat)
    u"\uFE3D": u"<<",  # PRESENTATION FORM FOR VERTICAL LEFT DOUBLE ANGLE BRACKET (compat)
    u"\uFE3E": u">>",  # PRESENTATION FORM FOR VERTICAL RIGHT DOUBLE ANGLE BRACKET (compat)
    u"\uFE3F": u"<",  # PRESENTATION FORM FOR VERTICAL LEFT ANGLE BRACKET (compat)
    u"\uFE40": u">",  # PRESENTATION FORM FOR VERTICAL RIGHT ANGLE BRACKET (compat)
    u"\uFE51": u",",  # SMALL IDEOGRAPHIC COMMA (compat)
    u"\uFE5D": u"[",  # SMALL LEFT TORTOISE SHELL BRACKET (compat)
    u"\uFE5E": u"]",  # SMALL RIGHT TORTOISE SHELL BRACKET (compat)
    u"\uFF5F": u"((",  # FULLWIDTH LEFT WHITE PARENTHESIS (compat)(from ‹character-fallback›)
    u"\uFF60": u"))",  # FULLWIDTH RIGHT WHITE PARENTHESIS (compat)(from ‹character-fallback›)
    u"\uFF61": u".",  # HALFWIDTH IDEOGRAPHIC FULL STOP (compat)
    u"\uFF64": u",",  # HALFWIDTH IDEOGRAPHIC COMMA (compat)
    u"\u2212": u"-",  # MINUS SIGN (from ‹character-fallback›)
    u"\u2215": u"/",  # DIVISION SLASH (from ‹character-fallback›)
    u"\u2216": u"\\",  # SET MINUS (from ‹character-fallback›)
    u"\u2223": u"|",  # DIVIDES (from ‹character-fallback›)
    u"\u2225": u"||",  # PARALLEL TO (from ‹character-fallback›)
    u"\u226A": u"<<",  # MUCH LESS-THAN
    u"\u226B": u">>",  # MUCH GREATER-THAN
    u"\u2985": u"((",  # LEFT WHITE PARENTHESIS
    u"\u2986": u"))",  # RIGHT WHITE PARENTHESIS
    u"\u200B": u"",  # Zero Width Space
}
_re_simplify_punctuation = _re_any(_simplify_punctuation.keys())


def unicode_simplify_punctuation(string):
    return _re_simplify_punctuation.sub(lambda m: _simplify_punctuation[m.group(0)], string)


_simplify_combinations = {
    u"\u00C6": u"AE",  # LATIN CAPITAL LETTER AE (from ‹character-fallback›)
    u"\u00D0": u"D",  # LATIN CAPITAL LETTER ETH
    u"\u00D8": u"OE",  # LATIN CAPITAL LETTER O WITH STROKE (see https://en.wikipedia.org/wiki/%C3%98)
    u"\u00DE": u"TH",  # LATIN CAPITAL LETTER THORN
    u"\u00DF": u"ss",  # LATIN SMALL LETTER SHARP S (from ‹character-fallback›)
    u"\u00E6": u"ae",  # LATIN SMALL LETTER AE (from ‹character-fallback›)
    u"\u00F0": u"d",  # LATIN SMALL LETTER ETH
    u"\u00F8": u"oe",  # LATIN SMALL LETTER O WITH STROKE (see https://en.wikipedia.org/wiki/%C3%98)
    u"\u00FE": u"th",  # LATIN SMALL LETTER THORN
    u"\u0110": u"D",  # LATIN CAPITAL LETTER D WITH STROKE
    u"\u0111": u"d",  # LATIN SMALL LETTER D WITH STROKE
    u"\u0126": u"H",  # LATIN CAPITAL LETTER H WITH STROKE
    u"\u0127": u"h",  # LATIN CAPITAL LETTER H WITH STROKE
    u"\u0131": u"i",  # LATIN SMALL LETTER DOTLESS I
    u"\u0138": u"q",  # LATIN SMALL LETTER KRA (collates with q in DUCET)
    u"\u0141": u"L",  # LATIN CAPITAL LETTER L WITH STROKE
    u"\u0142": u"l",  # LATIN SMALL LETTER L WITH STROKE
    u"\u0149": u"'n",  # LATIN SMALL LETTER N PRECEDED BY APOSTROPHE (from ‹character-fallback›)
    u"\u014A": u"N",  # LATIN CAPITAL LETTER ENG
    u"\u014B": u"n",  # LATIN SMALL LETTER ENG
    u"\u0152": u"OE",  # LATIN CAPITAL LIGATURE OE (from ‹character-fallback›)
    u"\u0153": u"oe",  # LATIN SMALL LIGATURE OE (from ‹character-fallback›)
    u"\u0166": u"T",  # LATIN CAPITAL LETTER T WITH STROKE
    u"\u0167": u"t",  # LATIN SMALL LETTER T WITH STROKE
    u"\u0180": u"b",  # LATIN SMALL LETTER B WITH STROKE
    u"\u0181": u"B",  # LATIN CAPITAL LETTER B WITH HOOK
    u"\u0182": u"B",  # LATIN CAPITAL LETTER B WITH TOPBAR
    u"\u0183": u"b",  # LATIN SMALL LETTER B WITH TOPBAR
    u"\u0187": u"C",  # LATIN CAPITAL LETTER C WITH HOOK
    u"\u0188": u"c",  # LATIN SMALL LETTER C WITH HOOK
    u"\u0189": u"D",  # LATIN CAPITAL LETTER AFRICAN D
    u"\u018A": u"D",  # LATIN CAPITAL LETTER D WITH HOOK
    u"\u018B": u"D",  # LATIN CAPITAL LETTER D WITH TOPBAR
    u"\u018C": u"d",  # LATIN SMALL LETTER D WITH TOPBAR
    u"\u0190": u"E",  # LATIN CAPITAL LETTER OPEN E
    u"\u0191": u"F",  # LATIN CAPITAL LETTER F WITH HOOK
    u"\u0192": u"f",  # LATIN SMALL LETTER F WITH HOOK
    u"\u0193": u"G",  # LATIN CAPITAL LETTER G WITH HOOK
    u"\u0195": u"hv",  # LATIN SMALL LETTER HV
    u"\u0196": u"I",  # LATIN CAPITAL LETTER IOTA
    u"\u0197": u"I",  # LATIN CAPITAL LETTER I WITH STROKE
    u"\u0198": u"K",  # LATIN CAPITAL LETTER K WITH HOOK
    u"\u0199": u"k",  # LATIN SMALL LETTER K WITH HOOK
    u"\u019A": u"l",  # LATIN SMALL LETTER L WITH BAR
    u"\u019D": u"N",  # LATIN CAPITAL LETTER N WITH LEFT HOOK
    u"\u019E": u"n",  # LATIN SMALL LETTER N WITH LONG RIGHT LEG
    u"\u01A2": u"GH",  # LATIN CAPITAL LETTER GHA (see http://unicode.org/notes/tn27/)
    u"\u01A3": u"gh",  # LATIN SMALL LETTER GHA (see http://unicode.org/notes/tn27/)
    u"\u01A4": u"P",  # LATIN CAPITAL LETTER P WITH HOOK
    u"\u01A5": u"p",  # LATIN SMALL LETTER P WITH HOOK
    u"\u01AB": u"t",  # LATIN SMALL LETTER T WITH PALATAL HOOK
    u"\u01AC": u"T",  # LATIN CAPITAL LETTER T WITH HOOK
    u"\u01AD": u"t",  # LATIN SMALL LETTER T WITH HOOK
    u"\u01AE": u"T",  # LATIN CAPITAL LETTER T WITH RETROFLEX HOOK
    u"\u01B2": u"V",  # LATIN CAPITAL LETTER V WITH HOOK
    u"\u01B3": u"Y",  # LATIN CAPITAL LETTER Y WITH HOOK
    u"\u01B4": u"y",  # LATIN SMALL LETTER Y WITH HOOK
    u"\u01B5": u"Z",  # LATIN CAPITAL LETTER Z WITH STROKE
    u"\u01B6": u"z",  # LATIN SMALL LETTER Z WITH STROKE
    u"\u01C4": u"DZ",  # LATIN CAPITAL LETTER DZ WITH CARON (compat)
    u"\u01C5": u"Dz",  # LATIN CAPITAL LETTER D WITH SMALL LETTER Z WITH CARON (compat)
    u"\u01C6": u"dz",  # LATIN SMALL LETTER DZ WITH CARON (compat)
    u"\u01E4": u"G",  # LATIN CAPITAL LETTER G WITH STROKE
    u"\u01E5": u"g",  # LATIN SMALL LETTER G WITH STROKE
    u"\u0221": u"d",  # LATIN SMALL LETTER D WITH CURL
    u"\u0224": u"Z",  # LATIN CAPITAL LETTER Z WITH HOOK
    u"\u0225": u"z",  # LATIN SMALL LETTER Z WITH HOOK
    u"\u0234": u"l",  # LATIN SMALL LETTER L WITH CURL
    u"\u0235": u"n",  # LATIN SMALL LETTER N WITH CURL
    u"\u0236": u"t",  # LATIN SMALL LETTER T WITH CURL
    u"\u0237": u"j",  # LATIN SMALL LETTER DOTLESS J
    u"\u0238": u"db",  # LATIN SMALL LETTER DB DIGRAPH
    u"\u0239": u"qp",  # LATIN SMALL LETTER QP DIGRAPH
    u"\u023A": u"A",  # LATIN CAPITAL LETTER A WITH STROKE
    u"\u023B": u"C",  # LATIN CAPITAL LETTER C WITH STROKE
    u"\u023C": u"c",  # LATIN SMALL LETTER C WITH STROKE
    u"\u023D": u"L",  # LATIN CAPITAL LETTER L WITH BAR
    u"\u023E": u"T",  # LATIN CAPITAL LETTER T WITH DIAGONAL STROKE
    u"\u023F": u"s",  # LATIN SMALL LETTER S WITH SWASH TAIL
    u"\u0240": u"z",  # LATIN SMALL LETTER Z WITH SWASH TAIL
    u"\u0243": u"B",  # LATIN CAPITAL LETTER B WITH STROKE
    u"\u0244": u"U",  # LATIN CAPITAL LETTER U BAR
    u"\u0246": u"E",  # LATIN CAPITAL LETTER E WITH STROKE
    u"\u0247": u"e",  # LATIN SMALL LETTER E WITH STROKE
    u"\u0248": u"J",  # LATIN CAPITAL LETTER J WITH STROKE
    u"\u0249": u"j",  # LATIN SMALL LETTER J WITH STROKE
    u"\u024C": u"R",  # LATIN CAPITAL LETTER R WITH STROKE
    u"\u024D": u"r",  # LATIN SMALL LETTER R WITH STROKE
    u"\u024E": u"Y",  # LATIN CAPITAL LETTER Y WITH STROKE
    u"\u024F": u"y",  # LATIN SMALL LETTER Y WITH STROKE
    u"\u0253": u"b",  # LATIN SMALL LETTER B WITH HOOK
    u"\u0255": u"c",  # LATIN SMALL LETTER C WITH CURL
    u"\u0256": u"d",  # LATIN SMALL LETTER D WITH TAIL
    u"\u0257": u"d",  # LATIN SMALL LETTER D WITH HOOK
    u"\u025B": u"e",  # LATIN SMALL LETTER OPEN E
    u"\u025F": u"j",  # LATIN SMALL LETTER DOTLESS J WITH STROKE
    u"\u0260": u"g",  # LATIN SMALL LETTER G WITH HOOK
    u"\u0261": u"g",  # LATIN SMALL LETTER SCRIPT G
    u"\u0262": u"G",  # LATIN LETTER SMALL CAPITAL G
    u"\u0266": u"h",  # LATIN SMALL LETTER H WITH HOOK
    u"\u0267": u"h",  # LATIN SMALL LETTER HENG WITH HOOK
    u"\u0268": u"i",  # LATIN SMALL LETTER I WITH STROKE
    u"\u026A": u"I",  # LATIN LETTER SMALL CAPITAL I
    u"\u026B": u"l",  # LATIN SMALL LETTER L WITH MIDDLE TILDE
    u"\u026C": u"l",  # LATIN SMALL LETTER L WITH BELT
    u"\u026D": u"l",  # LATIN SMALL LETTER L WITH RETROFLEX HOOK
    u"\u0271": u"m",  # LATIN SMALL LETTER M WITH HOOK
    u"\u0272": u"n",  # LATIN SMALL LETTER N WITH LEFT HOOK
    u"\u0273": u"n",  # LATIN SMALL LETTER N WITH RETROFLEX HOOK
    u"\u0274": u"N",  # LATIN LETTER SMALL CAPITAL N
    u"\u0276": u"OE",  # LATIN LETTER SMALL CAPITAL OE
    u"\u027C": u"r",  # LATIN SMALL LETTER R WITH LONG LEG
    u"\u027D": u"r",  # LATIN SMALL LETTER R WITH TAIL
    u"\u027E": u"r",  # LATIN SMALL LETTER R WITH FISHHOOK
    u"\u0280": u"R",  # LATIN LETTER SMALL CAPITAL R
    u"\u0282": u"s",  # LATIN SMALL LETTER S WITH HOOK
    u"\u0288": u"t",  # LATIN SMALL LETTER T WITH RETROFLEX HOOK
    u"\u0289": u"u",  # LATIN SMALL LETTER U BAR
    u"\u028B": u"v",  # LATIN SMALL LETTER V WITH HOOK
    u"\u028F": u"Y",  # LATIN LETTER SMALL CAPITAL Y
    u"\u0290": u"z",  # LATIN SMALL LETTER Z WITH RETROFLEX HOOK
    u"\u0291": u"z",  # LATIN SMALL LETTER Z WITH CURL
    u"\u0299": u"B",  # LATIN LETTER SMALL CAPITAL B
    u"\u029B": u"G",  # LATIN LETTER SMALL CAPITAL G WITH HOOK
    u"\u029C": u"H",  # LATIN LETTER SMALL CAPITAL H
    u"\u029D": u"j",  # LATIN SMALL LETTER J WITH CROSSED-TAIL
    u"\u029F": u"L",  # LATIN LETTER SMALL CAPITAL L
    u"\u02A0": u"q",  # LATIN SMALL LETTER Q WITH HOOK
    u"\u02A3": u"dz",  # LATIN SMALL LETTER DZ DIGRAPH
    u"\u02A5": u"dz",  # LATIN SMALL LETTER DZ DIGRAPH WITH CURL
    u"\u02A6": u"ts",  # LATIN SMALL LETTER TS DIGRAPH
    u"\u02AA": u"ls",  # LATIN SMALL LETTER LS DIGRAPH
    u"\u02AB": u"lz",  # LATIN SMALL LETTER LZ DIGRAPH
    u"\u1D01": u"AE",  # LATIN LETTER SMALL CAPITAL AE
    u"\u1D03": u"B",  # LATIN LETTER SMALL CAPITAL BARRED B
    u"\u1D06": u"D",  # LATIN LETTER SMALL CAPITAL ETH
    u"\u1D0C": u"L",  # LATIN LETTER SMALL CAPITAL L WITH STROKE
    u"\u1D6B": u"ue",  # LATIN SMALL LETTER UE
    u"\u1D6C": u"b",  # LATIN SMALL LETTER B WITH MIDDLE TILDE
    u"\u1D6D": u"d",  # LATIN SMALL LETTER D WITH MIDDLE TILDE
    u"\u1D6E": u"f",  # LATIN SMALL LETTER F WITH MIDDLE TILDE
    u"\u1D6F": u"m",  # LATIN SMALL LETTER M WITH MIDDLE TILDE
    u"\u1D70": u"n",  # LATIN SMALL LETTER N WITH MIDDLE TILDE
    u"\u1D71": u"p",  # LATIN SMALL LETTER P WITH MIDDLE TILDE
    u"\u1D72": u"r",  # LATIN SMALL LETTER R WITH MIDDLE TILDE
    u"\u1D73": u"r",  # LATIN SMALL LETTER R WITH FISHHOOK AND MIDDLE TILDE
    u"\u1D74": u"s",  # LATIN SMALL LETTER S WITH MIDDLE TILDE
    u"\u1D75": u"t",  # LATIN SMALL LETTER T WITH MIDDLE TILDE
    u"\u1D76": u"z",  # LATIN SMALL LETTER Z WITH MIDDLE TILDE
    u"\u1D7A": u"th",  # LATIN SMALL LETTER TH WITH STRIKETHROUGH
    u"\u1D7B": u"I",  # LATIN SMALL CAPITAL LETTER I WITH STROKE
    u"\u1D7D": u"p",  # LATIN SMALL LETTER P WITH STROKE
    u"\u1D7E": u"U",  # LATIN SMALL CAPITAL LETTER U WITH STROKE
    u"\u1D80": u"b",  # LATIN SMALL LETTER B WITH PALATAL HOOK
    u"\u1D81": u"d",  # LATIN SMALL LETTER D WITH PALATAL HOOK
    u"\u1D82": u"f",  # LATIN SMALL LETTER F WITH PALATAL HOOK
    u"\u1D83": u"g",  # LATIN SMALL LETTER G WITH PALATAL HOOK
    u"\u1D84": u"k",  # LATIN SMALL LETTER K WITH PALATAL HOOK
    u"\u1D85": u"l",  # LATIN SMALL LETTER L WITH PALATAL HOOK
    u"\u1D86": u"m",  # LATIN SMALL LETTER M WITH PALATAL HOOK
    u"\u1D87": u"n",  # LATIN SMALL LETTER N WITH PALATAL HOOK
    u"\u1D88": u"p",  # LATIN SMALL LETTER P WITH PALATAL HOOK
    u"\u1D89": u"r",  # LATIN SMALL LETTER R WITH PALATAL HOOK
    u"\u1D8A": u"s",  # LATIN SMALL LETTER S WITH PALATAL HOOK
    u"\u1D8C": u"v",  # LATIN SMALL LETTER V WITH PALATAL HOOK
    u"\u1D8D": u"x",  # LATIN SMALL LETTER X WITH PALATAL HOOK
    u"\u1D8E": u"z",  # LATIN SMALL LETTER Z WITH PALATAL HOOK
    u"\u1D8F": u"a",  # LATIN SMALL LETTER A WITH RETROFLEX HOOK
    u"\u1D91": u"d",  # LATIN SMALL LETTER D WITH HOOK AND TAIL
    u"\u1D92": u"e",  # LATIN SMALL LETTER E WITH RETROFLEX HOOK
    u"\u1D93": u"e",  # LATIN SMALL LETTER OPEN E WITH RETROFLEX HOOK
    u"\u1D96": u"i",  # LATIN SMALL LETTER I WITH RETROFLEX HOOK
    u"\u1D99": u"u",  # LATIN SMALL LETTER U WITH RETROFLEX HOOK
    u"\u1E9A": u"a",  # LATIN SMALL LETTER A WITH RIGHT HALF RING
    u"\u1E9C": u"s",  # LATIN SMALL LETTER LONG S WITH DIAGONAL STROKE
    u"\u1E9D": u"s",  # LATIN SMALL LETTER LONG S WITH HIGH STROKE
    u"\u1E9E": u"SS",  # LATIN CAPITAL LETTER SHARP S
    u"\u1EFA": u"LL",  # LATIN CAPITAL LETTER MIDDLE-WELSH LL
    u"\u1EFB": u"ll",  # LATIN SMALL LETTER MIDDLE-WELSH LL
    u"\u1EFC": u"V",  # LATIN CAPITAL LETTER MIDDLE-WELSH V
    u"\u1EFD": u"v",  # LATIN SMALL LETTER MIDDLE-WELSH V
    u"\u1EFE": u"Y",  # LATIN CAPITAL LETTER Y WITH LOOP
    u"\u1EFF": u"y",  # LATIN SMALL LETTER Y WITH LOOP
    u"\u00A9": u"(C)",  # COPYRIGHT SIGN (from ‹character-fallback›)
    u"\u00AE": u"(R)",  # REGISTERED SIGN (from ‹character-fallback›)
    u"\u20A0": u"CE",  # EURO-CURRENCY SIGN (from ‹character-fallback›)
    u"\u20A2": u"Cr",  # CRUZEIRO SIGN (from ‹character-fallback›)
    u"\u20A3": u"Fr.",  # FRENCH FRANC SIGN (from ‹character-fallback›)
    u"\u20A4": u"L.",  # LIRA SIGN (from ‹character-fallback›)
    u"\u20A7": u"Pts",  # PESETA SIGN (from ‹character-fallback›)
    u"\u20BA": u"TL",  # TURKISH LIRA SIGN (from ‹character-fallback›)
    u"\u20B9": u"Rs",  # INDIAN RUPEE SIGN (from ‹character-fallback›)
    u"\u211E": u"Rx",  # PRESCRIPTION TAKE (from ‹character-fallback›)
    u"\u33A7": u"m/s",  # SQUARE M OVER S (compat) (from ‹character-fallback›)
    u"\u33AE": u"rad/s",  # SQUARE RAD OVER S (compat) (from ‹character-fallback›)
    u"\u33C6": u"C/kg",  # SQUARE C OVER KG (compat) (from ‹character-fallback›)
    u"\u33DE": u"V/m",  # SQUARE V OVER M (compat) (from ‹character-fallback›)
    u"\u33DF": u"A/m",  # SQUARE A OVER M (compat) (from ‹character-fallback›)
    u"\u00BC": u" 1/4",  # VULGAR FRACTION ONE QUARTER (from ‹character-fallback›)
    u"\u00BD": u" 1/2",  # VULGAR FRACTION ONE HALF (from ‹character-fallback›)
    u"\u00BE": u" 3/4",  # VULGAR FRACTION THREE QUARTERS (from ‹character-fallback›)
    u"\u2153": u" 1/3",  # VULGAR FRACTION ONE THIRD (from ‹character-fallback›)
    u"\u2154": u" 2/3",  # VULGAR FRACTION TWO THIRDS (from ‹character-fallback›)
    u"\u2155": u" 1/5",  # VULGAR FRACTION ONE FIFTH (from ‹character-fallback›)
    u"\u2156": u" 2/5",  # VULGAR FRACTION TWO FIFTHS (from ‹character-fallback›)
    u"\u2157": u" 3/5",  # VULGAR FRACTION THREE FIFTHS (from ‹character-fallback›)
    u"\u2158": u" 4/5",  # VULGAR FRACTION FOUR FIFTHS (from ‹character-fallback›)
    u"\u2159": u" 1/6",  # VULGAR FRACTION ONE SIXTH (from ‹character-fallback›)
    u"\u215A": u" 5/6",  # VULGAR FRACTION FIVE SIXTHS (from ‹character-fallback›)
    u"\u215B": u" 1/8",  # VULGAR FRACTION ONE EIGHTH (from ‹character-fallback›)
    u"\u215C": u" 3/8",  # VULGAR FRACTION THREE EIGHTHS (from ‹character-fallback›)
    u"\u215D": u" 5/8",  # VULGAR FRACTION FIVE EIGHTHS (from ‹character-fallback›)
    u"\u215E": u" 7/8",  # VULGAR FRACTION SEVEN EIGHTHS (from ‹character-fallback›)
    u"\u215F": u" 1/",  # FRACTION NUMERATOR ONE (from ‹character-fallback›)
    u"\u3001": u",",  # IDEOGRAPHIC COMMA
    u"\u3002": u".",  # IDEOGRAPHIC FULL STOP
    u"\u00D7": u"x",  # MULTIPLICATION SIGN
    u"\u00F7": u"/",  # DIVISION SIGN
    u"\u00B7": u".",  # MIDDLE DOT
    u"\u1E9F": u"dd",  # LATIN SMALL LETTER DELTA
    u"\u0184": u"H",  # LATIN CAPITAL LETTER TONE SIX
    u"\u0185": u"h",  # LATIN SMALL LETTER TONE SIX
    u"\u01BE": u"ts",  # LATIN LETTER TS LIGATION (see http://unicode.org/notes/tn27/)
}
_re_simplify_combinations = _re_any(_simplify_combinations)


def unicode_simplify_combinations(string):
    return _re_simplify_combinations.sub(lambda m: _simplify_combinations[m.group(0)], string)


def unicode_simplify_accents(string):
    result = ''.join(c for c in unicodedata.normalize('NFKD', string) if not unicodedata.combining(c))
    return result


def asciipunct(string):
    interim = unicode_simplify_compatibility(string)
    return unicode_simplify_punctuation(interim)


def unaccent(string):
    """Remove accents ``string``."""
    return unicode_simplify_accents(string)


def replace_non_ascii(string, repl="_"):
    """Replace non-ASCII characters from ``string`` by ``repl``."""
    interim = unicode_simplify_combinations(string)
    interim = unicode_simplify_accents(interim)
    interim = unicode_simplify_punctuation(interim)
    interim = unicode_simplify_compatibility(interim)

    def error_repl(e, repl=u"_"):
        return(repl, e.start + 1)
    codecs.register_error('repl', partial(error_repl, repl=unicode(repl)))

    return interim.encode('ascii', 'repl')



# setup.py - kantag setup configuration.
# Copyright (C) 2018 David Gasaway
# https://bitbucket.org/dgasaway/kantag/

# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# see <http://www.gnu.org/licenses>.
from setuptools import setup
import os
import io
from kantag._version import __version__

# Read the long description from the README.
basedir = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(basedir, 'README.rst'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

setup(
    name='kantag',
    version=__version__,
    description='manage audio file metadata using external text files',
    long_description=long_description,
    author='David Gasaway',
    author_email='dave@gasaway.org',
    url='https://bitbucket.org/dgasaway/kantag',
    download_url='https://bitbucket.org/dgasaway/kantag/downloads/',
    license='GNU GPL v2',
    packages=['kantag'],
    scripts=['bin/applykan', 'bin/initkan', 'bin/showkan'],
    python_requires='~=2.7',
    install_requires=[
        'mutagen >= 1.20',
    ],
    extras_require={
        'MusicBrainz': ['musicbrainzngs >= 0.6'],
    },
    keywords='audio music metadata tags',
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
    ],
)
